# --------------------------------------------------------------------------------------------------
# Code taken from https://github.com/iwantooxxoox/Keras-OpenFace (with minor modifications)
# --------------------------------------------------------------------------------------------------

import tensorflow as tf
from keras import backend as K
from keras.layers import (
    Activation,
    AveragePooling2D,
    BatchNormalization,
    concatenate,
    Conv2D,
    Input,
    MaxPooling2D,
    ZeroPadding2D
)
from keras.layers.core import Lambda, Flatten, Dense
from keras.models import Model

from pytoolbox.ai.vision import utils

DEFAULT_WEIGHTS = (
    'https://s3-eu-west-1.amazonaws.com/pytoolbox/ai/vision/face/recognize/'
    'nn4.small2.v1.h5')

E = 0.00001  # Epsilon


def conv2d_bn(
    tensor,
    layer=None,
    cv1_out=None,
    cv1_filter=(1, 1),
    cv1_strides=(1, 1),
    cv2_out=None,
    cv2_filter=(3, 3),
    cv2_strides=(1, 1),
    padding=None,
):
    num = '' if cv2_out is None else '1'
    tensor = Conv2D(cv1_out, cv1_filter, strides=cv1_strides, name=layer + '_conv' + num)(tensor)
    tensor = BatchNormalization(axis=3, epsilon=E, name=layer + '_bn' + num)(tensor)
    tensor = Activation('relu')(tensor)
    if padding is None:
        return tensor
    tensor = ZeroPadding2D(padding=padding)(tensor)
    if cv2_out is None:
        return tensor
    tensor = Conv2D(cv2_out, cv2_filter, strides=cv2_strides, name=layer + '_conv2')(tensor)
    tensor = BatchNormalization(axis=3, epsilon=E, name=layer + '_bn2')(tensor)
    tensor = Activation('relu')(tensor)
    return tensor


def LRN2D(tensor):  # pylint:disable=invalid-name
    return tf.nn.lrn(tensor, alpha=1e-4, beta=0.75)


def create_model():  # pylint:disable=too-many-locals,too-many-statements

    inputs = Input(shape=(96, 96, 3))

    tensor = ZeroPadding2D(padding=(3, 3), input_shape=(96, 96, 3))(inputs)
    tensor = Conv2D(64, (7, 7), strides=(2, 2), name='conv1')(tensor)
    tensor = BatchNormalization(axis=3, epsilon=E, name='bn1')(tensor)
    tensor = Activation('relu')(tensor)
    tensor = ZeroPadding2D(padding=(1, 1))(tensor)
    tensor = MaxPooling2D(pool_size=3, strides=2)(tensor)
    tensor = Lambda(LRN2D, name='lrn_1')(tensor)
    tensor = Conv2D(64, (1, 1), name='conv2')(tensor)
    tensor = BatchNormalization(axis=3, epsilon=E, name='bn2')(tensor)
    tensor = Activation('relu')(tensor)
    tensor = ZeroPadding2D(padding=(1, 1))(tensor)
    tensor = Conv2D(192, (3, 3), name='conv3')(tensor)
    tensor = BatchNormalization(axis=3, epsilon=E, name='bn3')(tensor)
    tensor = Activation('relu')(tensor)
    tensor = Lambda(LRN2D, name='lrn_2')(tensor)
    tensor = ZeroPadding2D(padding=(1, 1))(tensor)
    tensor = MaxPooling2D(pool_size=3, strides=2)(tensor)

    # Inception 3a

    Norm = BatchNormalization

    inception_3a_3x3 = Conv2D(96, (1, 1), name='inception_3a_3x3_conv1')(tensor)
    inception_3a_3x3 = Norm(axis=3, epsilon=E, name='inception_3a_3x3_bn1')(inception_3a_3x3)
    inception_3a_3x3 = Activation('relu')(inception_3a_3x3)
    inception_3a_3x3 = ZeroPadding2D(padding=(1, 1))(inception_3a_3x3)
    inception_3a_3x3 = Conv2D(128, (3, 3), name='inception_3a_3x3_conv2')(inception_3a_3x3)
    inception_3a_3x3 = Norm(axis=3, epsilon=E, name='inception_3a_3x3_bn2')(inception_3a_3x3)
    inception_3a_3x3 = Activation('relu')(inception_3a_3x3)

    inception_3a_5x5 = Conv2D(16, (1, 1), name='inception_3a_5x5_conv1')(tensor)
    inception_3a_5x5 = Norm(axis=3, epsilon=E, name='inception_3a_5x5_bn1')(inception_3a_5x5)
    inception_3a_5x5 = Activation('relu')(inception_3a_5x5)
    inception_3a_5x5 = ZeroPadding2D(padding=(2, 2))(inception_3a_5x5)
    inception_3a_5x5 = Conv2D(32, (5, 5), name='inception_3a_5x5_conv2')(inception_3a_5x5)
    inception_3a_5x5 = Norm(axis=3, epsilon=E, name='inception_3a_5x5_bn2')(inception_3a_5x5)
    inception_3a_5x5 = Activation('relu')(inception_3a_5x5)

    inception_3a_pool = MaxPooling2D(pool_size=3, strides=2)(tensor)
    inception_3a_pool = Conv2D(32, (1, 1), name='inception_3a_pool_conv')(inception_3a_pool)
    inception_3a_pool = Norm(axis=3, epsilon=E, name='inception_3a_pool_bn')(inception_3a_pool)
    inception_3a_pool = Activation('relu')(inception_3a_pool)
    inception_3a_pool = ZeroPadding2D(padding=((3, 4), (3, 4)))(inception_3a_pool)

    inception_3a_1x1 = Conv2D(64, (1, 1), name='inception_3a_1x1_conv')(tensor)
    inception_3a_1x1 = Norm(axis=3, epsilon=E, name='inception_3a_1x1_bn')(inception_3a_1x1)
    inception_3a_1x1 = Activation('relu')(inception_3a_1x1)

    inception_3a = concatenate([
        inception_3a_3x3,
        inception_3a_5x5,
        inception_3a_pool,
        inception_3a_1x1
    ], axis=3)

    # Inception 3b

    inception_3b_3x3 = Conv2D(96, (1, 1), name='inception_3b_3x3_conv1')(inception_3a)
    inception_3b_3x3 = Norm(axis=3, epsilon=E, name='inception_3b_3x3_bn1')(inception_3b_3x3)
    inception_3b_3x3 = Activation('relu')(inception_3b_3x3)
    inception_3b_3x3 = ZeroPadding2D(padding=(1, 1))(inception_3b_3x3)
    inception_3b_3x3 = Conv2D(128, (3, 3), name='inception_3b_3x3_conv2')(inception_3b_3x3)
    inception_3b_3x3 = Norm(axis=3, epsilon=E, name='inception_3b_3x3_bn2')(inception_3b_3x3)
    inception_3b_3x3 = Activation('relu')(inception_3b_3x3)

    inception_3b_5x5 = Conv2D(32, (1, 1), name='inception_3b_5x5_conv1')(inception_3a)
    inception_3b_5x5 = Norm(axis=3, epsilon=E, name='inception_3b_5x5_bn1')(inception_3b_5x5)
    inception_3b_5x5 = Activation('relu')(inception_3b_5x5)
    inception_3b_5x5 = ZeroPadding2D(padding=(2, 2))(inception_3b_5x5)
    inception_3b_5x5 = Conv2D(64, (5, 5), name='inception_3b_5x5_conv2')(inception_3b_5x5)
    inception_3b_5x5 = Norm(axis=3, epsilon=E, name='inception_3b_5x5_bn2')(inception_3b_5x5)
    inception_3b_5x5 = Activation('relu')(inception_3b_5x5)

    inception_3b_pool = AveragePooling2D(pool_size=(3, 3), strides=(3, 3))(inception_3a)
    inception_3b_pool = Conv2D(64, (1, 1), name='inception_3b_pool_conv')(inception_3b_pool)
    inception_3b_pool = Norm(axis=3, epsilon=E, name='inception_3b_pool_bn')(inception_3b_pool)
    inception_3b_pool = Activation('relu')(inception_3b_pool)
    inception_3b_pool = ZeroPadding2D(padding=(4, 4))(inception_3b_pool)

    inception_3b_1x1 = Conv2D(64, (1, 1), name='inception_3b_1x1_conv')(inception_3a)
    inception_3b_1x1 = Norm(axis=3, epsilon=E, name='inception_3b_1x1_bn')(inception_3b_1x1)
    inception_3b_1x1 = Activation('relu')(inception_3b_1x1)

    inception_3b = concatenate([
        inception_3b_3x3,
        inception_3b_5x5,
        inception_3b_pool,
        inception_3b_1x1
    ], axis=3)

    # Inception 3c

    inception_3c_3x3 = conv2d_bn(
        inception_3b,
        layer='inception_3c_3x3',
        cv1_out=128,
        cv1_filter=(1, 1),
        cv2_out=256,
        cv2_filter=(3, 3),
        cv2_strides=(2, 2),
        padding=(1, 1))

    inception_3c_5x5 = conv2d_bn(
        inception_3b,
        layer='inception_3c_5x5',
        cv1_out=32,
        cv1_filter=(1, 1),
        cv2_out=64,
        cv2_filter=(5, 5),
        cv2_strides=(2, 2),
        padding=(2, 2))

    inception_3c_pool = MaxPooling2D(pool_size=3, strides=2)(inception_3b)
    inception_3c_pool = ZeroPadding2D(padding=((0, 1), (0, 1)))(inception_3c_pool)

    inception_3c = concatenate([
        inception_3c_3x3,
        inception_3c_5x5,
        inception_3c_pool
    ], axis=3)

    # Inception 4a

    inception_4a_3x3 = conv2d_bn(
        inception_3c,
        layer='inception_4a_3x3',
        cv1_out=96,
        cv1_filter=(1, 1),
        cv2_out=192,
        cv2_filter=(3, 3),
        cv2_strides=(1, 1),
        padding=(1, 1))

    inception_4a_5x5 = conv2d_bn(
        inception_3c,
        layer='inception_4a_5x5',
        cv1_out=32,
        cv1_filter=(1, 1),
        cv2_out=64,
        cv2_filter=(5, 5),
        cv2_strides=(1, 1),
        padding=(2, 2))

    inception_4a_pool = AveragePooling2D(pool_size=(3, 3), strides=(3, 3))(inception_3c)

    inception_4a_pool = conv2d_bn(
        inception_4a_pool,
        layer='inception_4a_pool',
        cv1_out=128,
        cv1_filter=(1, 1),
        padding=(2, 2))

    inception_4a_1x1 = conv2d_bn(
        inception_3c,
        layer='inception_4a_1x1',
        cv1_out=256,
        cv1_filter=(1, 1))

    inception_4a = concatenate([
        inception_4a_3x3,
        inception_4a_5x5,
        inception_4a_pool,
        inception_4a_1x1
    ], axis=3)

    # Inception 4e

    inception_4e_3x3 = conv2d_bn(
        inception_4a,
        layer='inception_4e_3x3',
        cv1_out=160,
        cv1_filter=(1, 1),
        cv2_out=256,
        cv2_filter=(3, 3),
        cv2_strides=(2, 2),
        padding=(1, 1))

    inception_4e_5x5 = conv2d_bn(
        inception_4a,
        layer='inception_4e_5x5',
        cv1_out=64,
        cv1_filter=(1, 1),
        cv2_out=128,
        cv2_filter=(5, 5),
        cv2_strides=(2, 2),
        padding=(2, 2))

    inception_4e_pool = MaxPooling2D(pool_size=3, strides=2)(inception_4a)
    inception_4e_pool = ZeroPadding2D(padding=((0, 1), (0, 1)))(inception_4e_pool)

    inception_4e = concatenate([
        inception_4e_3x3,
        inception_4e_5x5,
        inception_4e_pool
    ], axis=3)

    # Inception 5a

    inception_5a_3x3 = conv2d_bn(
        inception_4e,
        layer='inception_5a_3x3',
        cv1_out=96,
        cv1_filter=(1, 1),
        cv2_out=384,
        cv2_filter=(3, 3),
        cv2_strides=(1, 1),
        padding=(1, 1))

    inception_5a_pool = AveragePooling2D(pool_size=(3, 3), strides=(3, 3))(inception_4e)

    inception_5a_pool = conv2d_bn(
        inception_5a_pool,
        layer='inception_5a_pool',
        cv1_out=96,
        cv1_filter=(1, 1),
        padding=(1, 1))

    inception_5a_1x1 = conv2d_bn(
        inception_4e,
        layer='inception_5a_1x1',
        cv1_out=256,
        cv1_filter=(1, 1))

    inception_5a = concatenate([
        inception_5a_3x3,
        inception_5a_pool,
        inception_5a_1x1
    ], axis=3)

    # Inception 5b

    inception_5b_3x3 = conv2d_bn(
        inception_5a,
        layer='inception_5b_3x3',
        cv1_out=96,
        cv1_filter=(1, 1),
        cv2_out=384,
        cv2_filter=(3, 3),
        cv2_strides=(1, 1),
        padding=(1, 1))

    inception_5b_pool = MaxPooling2D(pool_size=3, strides=2)(inception_5a)

    inception_5b_pool = conv2d_bn(
        inception_5b_pool,
        layer='inception_5b_pool',
        cv1_out=96,
        cv1_filter=(1, 1))

    inception_5b_pool = ZeroPadding2D(padding=(1, 1))(inception_5b_pool)

    inception_5b_1x1 = conv2d_bn(
        inception_5a,
        layer='inception_5b_1x1',
        cv1_out=256,
        cv1_filter=(1, 1))

    inception_5b = concatenate([
        inception_5b_3x3,
        inception_5b_pool,
        inception_5b_1x1
    ], axis=3)

    # Outputs

    reshape_layer = Flatten()(AveragePooling2D(pool_size=(3, 3), strides=(1, 1))(inception_5b))
    dense_layer = Dense(128, name='dense_layer')(reshape_layer)

    outputs = Lambda(lambda x: K.l2_normalize(x, axis=1), name='norm_layer')(dense_layer)

    return Model(inputs=[inputs], outputs=outputs)


def load_model(weights=DEFAULT_WEIGHTS):
    model = create_model()
    if weights:
        model.load_weights(utils.load_to_file(weights))
    return model



from selenium.webdriver.support import select

__all__ = ['Select']


class Select(select.Select):
    """A Select with the attributes of the WebElement."""

    def __getattr__(self, name):
        return getattr(self._el, name)



import datetime, re
from fractions import Fraction

from pytoolbox import decorators
from pytoolbox.datetime import str_to_datetime, str_to_time
from . import brand  # pylint:disable=unused-import

__all__ = ['Tag', 'TagSet']


class Tag(object):

    brand_class = brand.Brand
    date_formats = ['%Y%m%d %H%M%S', '%Y%m%d']
    date_clean_regex = re.compile(r'[:-]')
    group_to_brand_blacklist = frozenset(['aux', 'crs', 'exifEX', 'Image', 'Photo'])
    type_to_hook = {
        datetime.datetime: 'get_tag_string',  # clean method will convert to a date-time
        datetime.time: 'get_tag_string',  # clean method will convert to a time
        Fraction: 'get_exif_tag_rational',
        int: 'get_tag_long',
        list: 'get_tag_multiple',
        str: 'get_tag_string'
    }
    type_to_python = {
        'Ascii': str,
        'Byte': bytes,
        'Comment': str,
        'Date': datetime.datetime,
        'LangAlt': str,
        'Long': int,
        'Rational': Fraction,
        'SRational': Fraction,
        'SLong': int,
        'Short': int,
        'SShort': int,
        'String': str,
        'Time': datetime.time,
        'Undefined': bytes,
        'XmpBag': list,
        'XmpSeq': list,
        'XmpText': str
    }

    def __init__(self, metadata, key):
        self.metadata = metadata
        self.key = key

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.key}: {str(self.data)[:20]}>'

    @property
    def data(self):
        type_hook = self.get_type_hook()
        if type_hook:
            try:
                return self.clean(type_hook(self.key))
            except UnicodeDecodeError as e:
                return e
        return self.data_bytes

    @property
    def data_bytes(self):
        tag_raw = self.metadata.exiv2.get_tag_raw(self.key)
        return tag_raw.get_data() if tag_raw else None

    @decorators.cached_property
    def description(self):
        return self.metadata.exiv2.get_tag_description(self.key)

    @property
    def brand(self):
        if self.group in self.group_to_brand_blacklist:
            return None
        return self.brand_class(self.group)

    @property
    def group(self):
        return self.key.split('.')[-2]

    @decorators.cached_property
    def label(self):
        return self.metadata.exiv2.get_tag_label(self.key) or self.key.split('.')[-1]

    @property
    def size(self):
        return self.metadata.exiv2.get_tag_raw(self.key).get_size()

    @decorators.cached_property
    def type(self):
        tag_type = self.metadata.exiv2.get_tag_type(self.key)
        try:
            return self.type_to_python[tag_type]
        except KeyError as e:
            if tag_type:
                raise KeyError(f'Unknow tag type {tag_type}') from e
        return bytes

    def clean(self, data):
        if isinstance(data, str):
            data = data.strip()
        if self.type == datetime.time:
            return str_to_time(data)
        if self.type == datetime.datetime or isinstance(data, str):
            cleaned_data = self.date_clean_regex.sub('', data)
            for date_format in self.date_formats:
                date = str_to_datetime(cleaned_data, date_format, fail=False)
                if date:
                    return date
        assert not data or isinstance(data, self.type), \
            f'{self.key} {self.type} {data} {type(data)}'
        return data

    def get_type_hook(self):
        name = self.type_to_hook.get(self.type)
        return getattr(self.metadata.exiv2, name) if name else None


class TagSet(object):

    @staticmethod
    def clean_number(number):
        return number if number and number > 0 else None

    def __init__(self, metadata):
        self.metadata = metadata



from .brand import *      # noqa:F401,F403
from .camera import *     # noqa:F401,F403
from .equipment import *  # noqa:F401,F403
from .image import *      # noqa:F401,F403
from .lens import *       # noqa:F401,F403
from .metadata import *   # noqa:F401,F403
from .photo import *      # noqa:F401,F403
from .tag import *        # noqa:F401,F403



"""
Mix-ins for building your own `Django Filter <https://github.com/alex/django-filter>`_
powered filters.
"""

__all__ = ['RaiseOnUnhandledFieldClassMixin']


class RaiseOnUnhandledFieldClassMixin(object):
    """
    Raise an exception when the filter set is unable to find a suitable filter for any of the model
    fields to filter.
    """

    @classmethod
    def filter_for_field(cls, f, name, lookup_type='exact'):
        value = super().filter_for_field(f, name, lookup_type)
        if not value:
            raise NotImplementedError(
                f"Unable to find a suitable filter for field '{name}' of class {f.__class__} "
                f"with lookup type '{lookup_type}'")
        return value



"""
Abstract models for building your own models.
"""

from django.db import models

from . import fields

__all__ = ['Timestamped']


class Timestamped(models.Model):

    class Meta:
        abstract = True
        get_latest_by = 'created_at'
        ordering = ('created_at', 'updated_at')

    created_at = fields.CreatedAtField()
    updated_at = fields.UpdatedAtField()



from selenium.common import exceptions

# Replace the infamous "from selenium.common.exceptions import *" :)
things = {k: v for k, v in exceptions.__dict__.items() if k[0] != '_'}
__all__ = sorted(list(things.keys()) + ['NoSuchSpecializedElementException'])
assert 'NoSuchSpecializedElementException' not in things

globals().update(things)


class NoSuchSpecializedElementException(exceptions.NoSuchElementException):
    pass



"""
Extra forms.
"""

from django.utils.functional import cached_property

from pytoolbox.django.models import utils

__all__ = ['SerializedInstanceForm']


class SerializedInstanceForm(object):

    def __init__(self, **kwargs):
        self.app_label = kwargs['app_label']
        self.model = kwargs['model']
        self.pk = kwargs['pk']

    @classmethod
    def serialize(cls, instance):
        return utils.get_content_type_dict(instance)

    @cached_property
    def instance(self):
        return utils.get_instance(self.app_label, self.model, self.pk)

    def is_valid(self):
        try:
            return bool(self.instance)
        except Exception:
            return False



"""
Extra views.
"""

from django.http import HttpResponseRedirect
from django.views.generic.edit import DeleteView

__all__ = ['CancellableDeleteView']


class CancellableDeleteView(DeleteView):
    """Handle the cancel action (detect a cancel parameter in the POST request)."""

    def post(self, request, *args, **kwargs):
        if 'cancel' in request.POST:
            return HttpResponseRedirect(self.success_url)
        return super().post(request, *args, **kwargs)



"""
Module containing extensions to validate data with
`voluptuous <https://github.com/alecthomas/voluptuous>`_.
"""

import functools, re

import voluptuous

from . import module, validation

_all = module.All(globals())


# Errors

class PasswordInvalid(voluptuous.Invalid):
    """Incorrect password"""


class VersionInvalid(voluptuous.Invalid):
    """Incorrect version number"""


# Validators

@voluptuous.message('Incorrect e-mail address')
def Email(value):  # pylint:disable=invalid-name
    value = str(value)
    if not validation.valid_email(value):
        raise ValueError
    return value


@voluptuous.message('Incorrect list of e-mail addresses')
def EmailSet(values):  # pylint:disable=invalid-name
    emails = set()
    for value in values or []:
        email = str(value)
        if not validation.valid_email(email):
            raise ValueError
        emails.add(email)
    return emails


@voluptuous.message('Incorrect git commit hash')
def GitCommitHash(value):  # pylint:disable=invalid-name
    if re.match(r'^[0-9a-f]{40}$', value):
        return value
    raise ValueError


def Password(length=16, msg=None):  # pylint:disable=invalid-name

    @functools.wraps(Password)
    def f(value):
        value = str(value)
        if len(value) < length:
            raise PasswordInvalid(msg or 'Incorrect password')
    return f


@voluptuous.message('Incorrect percentage')
def Percent(value):  # pylint:disable=invalid-name
    return voluptuous.Range(min=1, max=100)(value)


@voluptuous.message('Incorrect SHA-256 checksum')
def SHA256(value):  # pylint:disable=invalid-name
    if re.match(r'^[0-9a-f]{64}$', value):
        return value
    raise ValueError


def Version(digits=4, msg=None):  # pylint:disable=invalid-name
    assert 1 <= digits <= 4
    pattern = r'^[0-9]+(\.[0-9]+){%d}\.[a-z0-9]+$'

    @functools.wraps(Version)
    def f(value):
        if re.match(pattern % (digits - 2), value):  # pylint:disable=consider-using-f-string
            return value
        raise VersionInvalid(msg or 'Incorrect version number')
    return f


__all__ = _all.diff(globals())



import itertools

from django.db import DatabaseError
from django.utils.translation import gettext_lazy as _

from pytoolbox import exceptions, module

_all = module.All(globals())


def get_message(validation_error):
    message, params = validation_error.message, validation_error.params
    return message % params if params else message


def has_code(validation_error, code):
    """
    **Example usage**

    >>> from django.core.exceptions import ValidationError
    >>> has_code(ValidationError('yo'), 'bad')
    False
    >>> has_code(ValidationError('yo', code='bad'), 'bad')
    True
    >>> has_code(ValidationError({'__all__': ValidationError('yo')}), 'bad')
    False
    >>> has_code(ValidationError({'__all__': ValidationError('yo', code='bad')}), 'bad')
    True
    >>> has_code(ValidationError([ValidationError('yo')]), 'bad')
    False
    >>> has_code(ValidationError([ValidationError('yo', code='bad')]), 'bad')
    True
    """
    errors = getattr(validation_error, 'error_list', [])
    errors_dicts = getattr(validation_error, 'error_dict', {})
    errors.extend(itertools.chain.from_iterable(errors_dicts.values()))
    return any(e.code == code for e in errors)


def iter_validation_errors(validation_error):
    """
    **Example usage**

    >>> from django.core.exceptions import ValidationError
    >>> from pytoolbox.unittest import asserts
    >>> eq = lambda i, l: asserts.list_equal(list(i), l)
    >>> bad, boy = ValidationError('yo', code='bad'), ValidationError('yo', code='boy')
    >>> eq(iter_validation_errors(bad), [(None, bad)])
    >>> eq(iter_validation_errors(ValidationError({'__all__': boy})), [('__all__', boy)])
    >>> eq(iter_validation_errors(ValidationError([bad, boy])), [(None, bad), (None, boy)])
    """
    if hasattr(validation_error, 'error_dict'):
        for field, errors in validation_error.error_dict.items():
            for error in errors:
                yield field, error
    else:
        for error in validation_error.error_list:
            yield None, error


class DatabaseUpdatePreconditionsError(exceptions.MessageMixin, DatabaseError):
    message = _('Row update request preconditions failed: '
                'A concurrent request changed the row in database.')


class InvalidStateError(exceptions.MessageMixin, Exception):
    message = _('State of {instance} is {instance.state}, excepted in any of {states}.')


class TransitionNotAllowedError(exceptions.MessageMixin, Exception):
    message = _('Cannot change state of {instance} from {instance.state} to {state}.')


__all__ = _all.diff(globals())



"""
Extra widgets for your forms.
"""

from django.forms import widgets
from django.utils.html import mark_safe

__all__ = ['CalendarDateInput', 'ClockTimeInput']


class CalendarDateInput(widgets.DateInput):
    def render(self, *args, **kwargs):
        html = super().render(*args, **kwargs)
        return mark_safe(
            '<div class="input-append date">'
            f'{html}<span class="add-on"><i class="icon-calendar"></i></span></div>')


class ClockTimeInput(widgets.TimeInput):
    def render(self, *args, **kwargs):
        html = super().render(*args, **kwargs)
        return mark_safe(
            '<div class="input-append bootstrap-timepicker">'
            f'{html}<span class="add-on"><i class="icon-time"></i></span></div>')



import re

from .ffmpeg import FFmpeg

__all__ = ['ENCODING_REGEX', 'X264']

# [79.5%] 3276/4123 frames, 284.69 fps, 2111.44 kb/s, eta 0:00:02
ENCODING_REGEX = re.compile(
    r'\[(?P<percent>\d+\.\d*)%\]\s+(?P<frame>\d+)/(?P<frame_total>\d+)\s+frames,\s+'
    r'(?P<frame_rate>\d+\.\d*)\s+fps,\s+(?P<bit_rate>[^,]+),\s+eta\s+(?P<eta>[\d:]+)'
)


class X264(FFmpeg):

    # encoding_regex = ENCODING_REGEX
    executable = 'x264'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        raise NotImplementedError('Must be reimplemented based on newer FFmpeg class interface')

    def _get_arguments(self, inputs, outputs, in_options=None, out_options=None):
        raise NotImplementedError('Must be reimplemented based on newer FFmpeg class interface')
        # in_paths = [f for f in ([in_paths] if isinstance(in_paths, str) else in_paths)]
        # if len(in_paths) > 1:
        #     raise NotImplementedError('Unable to handle more than one input.')
        # out_path = out_path or '/dev/null'
        # options = (shlex.split(options) if isinstance(options, str) else options) or []
        # args = [self.executable] + options + ['-o', out_path] + in_paths
        # return args, in_paths, out_path, options

    # def _get_progress(self, in_duration, stats):
    #     out_duration = in_duration * float(stats['percent'])
    #     ratio = float(stats['frame']) / float(stats['frame_total'])
    #     return out_duration, ratio

    # def _clean_statistics(self, stats, **statistics):
    #     statistics.setdefault('eta_time', datetime.timedelta(seconds=total_seconds(stats['eta'])))
    #     return statistics



from typing import Iterator, Optional
import difflib, operator as op, os
from packaging.version import parse as parse_version, LegacyVersion, Version

import termcolor

from .types import get_slots

__all__ = [
    'SlotsEqualityMixin',
    'unified_diff',
    'VERSION_OPERATIONS',
    'LegacyVersion',
    'Version',
    'compare_versions',
    'parse_version',
    'satisfy_version_constraints'
]


class SlotsEqualityMixin(object):
    """
    Implement the comparison operators based on the slots.
    Both the name of the slots retrieved with :func:`pytoolbox.types.get_slots`
    and theirs values are tested for equality.
    """

    def __eq__(self, other):
        return get_slots(self) == get_slots(other) and \
            all(getattr(self, a) == getattr(other, a) for a in get_slots(self))

    def __ne__(self, other):
        return not self.__eq__(other)


# Content ------------------------------------------------------------------------------------------

def unified_diff(before: str, after: str, *, colorize: bool = True, **kwargs) -> str:
    diff = difflib.unified_diff(before.splitlines(), after.splitlines(), **kwargs)
    return os.linesep.join(_colorize(diff) if colorize else diff)


def _colorize(diff: Iterator[str]) -> Iterator[str]:
    for line in diff:
        if line.startswith('+'):
            yield termcolor.colored(line, 'green')
        elif line.startswith('-'):
            yield termcolor.colored(line, 'red')
        elif line.startswith('^'):
            yield termcolor.colored(line, 'blue')
        else:
            yield line


# Versions -----------------------------------------------------------------------------------------

def _eqn(a, b) -> Optional[bool]:  # pylint:disable=invalid-name
    return True if a == b else None


def _nen(a, b) -> Optional[bool]:  # pylint:disable=invalid-name
    return False if a == b else None


VERSION_OPERATIONS = {
    Version: {'<': op.lt, '<=': op.le, '==': op.eq, '!=': op.ne, '>=': op.ge, '>': op.gt},
    LegacyVersion: {'<': _nen, '<=': _eqn, '==': op.eq, '!=': op.ne, '>=': _eqn, '>': _nen}
}


def compare_versions(
    a: str,  # pylint:disable=invalid-name
    b: str,  # pylint:disable=invalid-name
    operator: str
) -> Optional[bool]:
    version_a = parse_version(a)
    version_b = parse_version(b)
    if type(version_a) is type(version_b):
        operation = VERSION_OPERATIONS[type(version_a)][operator]
        return operation(version_a, version_b) if operation else None
    return None  # Will not try to compare Version to LegacyVersion


def satisfy_version_constraints(
    version: Optional[str],
    constraints: tuple[str, ...], *,
    default='<undefined>',
) -> bool:
    """
    Ensure given version fulfill the constraints (if any).

    Constraints are given in the form '<operator> <version>', Exemple:

    >>> satisfy_version_constraints('v1.5.2', ['>= v1.5', '< v2'])
    True
    >>> satisfy_version_constraints('v0.7', ['>= v1.5', '< v2'])
    False
    >>> satisfy_version_constraints(None, ['>= v1.5', '< v2'])
    False
    >>> satisfy_version_constraints('main', ['!= main'])
    False
    >>> satisfy_version_constraints(None, ['== <undefined>'])
    True
    >>> satisfy_version_constraints(None, ['!= master'], default='master')
    False
    """
    return all(compare_versions(version or default, *c.split(' ')[::-1]) for c in constraints or [])



from . import client  # pylint:disable=unused-import

__all__ = ['LiveTestCaseMixin']


class LiveTestCaseMixin(object):

    live_client_class = client.LiveClient

    def setUp(self):  # pylint:disable=invalid-name
        """Call super's setUp and instantiate a live test client, only once."""
        super().setUp()
        if not hasattr(self.__class__, 'client'):
            self.__class__.client = self.live_client_class(self.live_server_url)
        self.client = self.__class__.client

    @classmethod
    def tearDownClass(cls):  # pylint:disable=invalid-name
        """Quit the live-test client and call super's tearDownClass."""
        if hasattr(cls, 'client'):
            cls.client.quit()
        super().tearDownClass()

    # Asserts

    def assertElementEqual(self, name, value, enabled=True):  # pylint:disable=invalid-name
        """Check the properties of an element. Works with both WebElement and Select."""
        if enabled:
            self.assertElementIsEnabled(name)
        else:
            self.assertElementIsDisabled(name)
        element = self.client.find_name(name)
        Select = self.client.web_driver.web_element_classes['select']  # pylint:disable=invalid-name
        if isinstance(element, Select):
            self.assertSelectOptions(name, value)
        else:
            self.assertElementValue(name, value)

    def assertElementIsDisabled(self, name, *args, **kwargs):  # pylint:disable=invalid-name
        self.assertFalse(self.client.find_name(name).is_enabled(), *args, **kwargs)

    def assertElementIsEnabled(self, name, *args, **kwargs):  # pylint:disable=invalid-name
        self.assertTrue(self.client.find_name(name).is_enabled(), *args, **kwargs)

    def assertElementIsReadOnly(self, name):  # pylint:disable=invalid-name
        self.assertIsNotNone(self.client.find_name(name).get_attribute('readonly'))

    def assertElementValue(self, name, value, *args, **kwargs):  # pylint:disable=invalid-name
        element = self.client.find_name(name)
        operator = kwargs.pop('operator', lambda x: x)
        self.assertEqual(
            operator(element.get_attribute('value')),
            element.clean_value(value),
            *args,
            **kwargs)

    def assertSelectOptions(self, name, texts, *args, **kwargs):  # pylint:disable=invalid-name
        self.assertListEqual(
            sorted(o.text for o in self.client.find_name(name).all_selected_options),
            sorted([texts] if isinstance(texts, str) else texts), *args, **kwargs)



from io import BytesIO

from botocore.exceptions import ClientError

from pytoolbox.regex import from_path_patterns


def copy_object(s3, bucket_name, source_key, target_key):
    return s3.copy_object(
        CopySource={'Bucket': bucket_name, 'Key': source_key},
        Bucket=bucket_name,
        Key=target_key)


def get_bucket_location(s3, bucket_name):
    return s3.get_bucket_location(Bucket=bucket_name)['LocationConstraint']


def get_object_url(bucket_name, location, key):
    return f'https://s3-{location}.amazonaws.com/{bucket_name}/{key}'


def list_objects(s3, bucket_name, prefix='', patterns='*', regex=False):
    if prefix and prefix[-1] != '/':
        prefix += '/'
    patterns = from_path_patterns(patterns, regex=regex)
    for page in s3.get_paginator('list_objects').paginate(Bucket=bucket_name, Prefix=prefix):
        try:
            objects = page['Contents']
        except KeyError:
            return
        for obj in objects:
            key = obj['Key']
            if any(p.match(key) for p in patterns):
                yield obj


def load_object_meta(s3, bucket_name, path, fail=True):
    try:
        return s3.head_object(Bucket=bucket_name, Key=path)
    except ClientError as e:
        if 'Not Found' in str(e) and not fail:
            return None
        raise


def read_object(s3, bucket_name, path, file=None, fail=True):
    try:
        if file is None:
            with BytesIO() as f:
                s3.download_fileobj(bucket_name, path, f)
                return f.getvalue()
        else:
            s3.download_fileobj(bucket_name, path, file)
            file.seek(0)
            return file
    except ClientError as e:
        if 'Not Found' in str(e) and not fail:
            return None
        raise


def remove_objects(
    s3,
    bucket_name,
    prefix='',
    patterns=r'*',
    callback=lambda obj: True,
    regex=False,
    simulate=False
):
    """
    Remove objects matching pattern, by chunks of 1000 to be efficient.

    * Set `callback` to a function returning True if given object has to be removed.
    """
    objects = []
    for obj in list_objects(s3, bucket_name, prefix, patterns, regex=regex):
        if callback(obj):
            key = obj['Key']
            objects.append({'Key': key})
            yield obj
            if len(objects) == 1000 and not simulate:
                s3.delete_objects(Bucket=bucket_name, Delete={'Objects': objects})
                objects = []
    # Remove remaining objects
    if objects and not simulate:
        s3.delete_objects(Bucket=bucket_name, Delete={'Objects': objects})


def write_object(s3, bucket_name, path, file):
    s3.upload_fileobj(file, bucket_name, path)



"""
Extra fields for your forms.
"""

import copy

from django import forms

__all__ = ['StripCharField']


class StripCharField(forms.RegexField):
    default_widget_attrs = {'autofocus': 'autofocus'}
    max_length = None

    def __init__(self, **kwargs):
        attrs = copy.deepcopy(self.default_widget_attrs)
        attrs.update(kwargs.pop('widget_attrs', None) or {})
        super().__init__(
            r'\S+',
            max_length=self.max_length,
            widget=forms.TextInput(attrs),
            **kwargs)



from pytoolbox.network.rtp import RtpPacket
from .base import FecPacket

__all__ = ['FecGenerator']


class FecGenerator(object):  # pylint:disable=too-many-instance-attributes
    """
    A SMPTE 2022-1 FEC streams generator.
    This generator accept incoming RTP media packets and compute corresponding FEC packets.
    """

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @property
    def L(self):  # pylint:disable=invalid-name
        """
        Returns the Horizontal size of the FEC matrix (columns).

        **Example usage**

        >>> print(FecGenerator(4, 5).L)
        4
        """
        return self._L

    @property
    def D(self):  # pylint:disable=invalid-name
        """
        Returns the vertical size of the FEC matrix (rows).

        **Example usage**

        >>> print(FecGenerator(4, 5).D)
        5
        """
        return self._D

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructor >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    def __init__(self, L, D):  # pylint:disable=invalid-name,too-many-instance-attributes
        """
        Construct a FecGenerator.

        :param L: Horizontal size of the FEC matrix (columns)
        :type L: int
        :param D: Vertical size of the FEC matrix (rows)
        :type D: int
        """
        self._L, self._D = L, D  # pylint:disable=invalid-name
        self._col_sequence = self._row_sequence = 1
        self._media_sequence = None
        self._medias = []
        self._invalid = self._total = 0

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @staticmethod
    def on_new_col(col):
        """
        Called by FecGenerator when a new column FEC packet is generated and available for output.

        By default this method only print a message to `stdout`.

        .. seealso::

            You can `monkey patch <http://stackoverflow.com/questions/5626193>`_ it.

        :param col: Generated column FEC packet
        :type col: FecPacket
        :param caller: The generator that fired this method / event
        :type caller: FecGenerator
        """
        print(
            f'New COL FEC packet '
            f'seq={col.sequence} '
            f'snbase={col.snbase} '
            f'LxD={col.L}x{col.D} '
            f'trec={col.timestamp_recovery}')

    @staticmethod
    def on_new_row(row):
        """
        Called by FecGenerator when a new row FEC packet is generated and available for output.

        By default this method only print a message to `stdout`.

        .. seealso::

            You can `monkey patch <http://stackoverflow.com/questions/5626193>`_ it.

        :param row: Generated row FEC packet
        :type row: FecPacket
        :param caller: The generator that fired this method / event
        :type caller: FecGenerator
        """
        print(
            f'New ROW FEC packet '
            f'seq={row.sequence} '
            f'snbase={row.snbase} '
            f'LxD={row.L}x{row.D} '
            f'trec={row.timestamp_recovery}')

    def on_reset(self, media):
        """
        Called by FecGenerator when the algorithm is reseted (an incoming media is out of sequence).

        By default this method only print a message to `stdout`.

        .. seealso::

            You can `monkey patch <http://stackoverflow.com/questions/5626193>`_ it.

        :param media: Out of sequence media packet
        :type row: RtpPacket
        :param caller: The generator that fired this method / event
        :type caller: FecGenerator
        """
        print(
            f'Media seq={media.sequence} is out of sequence '
            f'(expected {self._media_sequence}) : FEC algorithm reseted !')

    def put_media(self, media):
        """
        Put an incoming media packet.

        :param media: Incoming media packet
        :type media: RtpPacket

        **Example usage**

        Testing input of out of sequence medias:

        >>> g = FecGenerator(4, 5)
        >>> R = RtpPacket
        >>> g.put_media(R.create(1, 100, R.MP2T_PT, bytearray('Tabby', 'utf-8')))
        Media seq=1 is out of sequence (expected None) : FEC algorithm reseted !
        >>> g.put_media(R.create(1, 100, R.MP2T_PT, bytearray('1234', 'utf-8')))
        Media seq=1 is out of sequence (expected 2) : FEC algorithm reseted !
        >>> g.put_media(R.create(4, 400, R.MP2T_PT, bytearray('abcd', 'utf-8')))
        Media seq=4 is out of sequence (expected 2) : FEC algorithm reseted !
        >>> g.put_media(R.create(2, 200, R.MP2T_PT, bytearray('python', 'utf-8')))
        Media seq=2 is out of sequence (expected 5) : FEC algorithm reseted !
        >>> g.put_media(R.create(2, 200, R.MP2T_PT, bytearray('Kuota Kharma Evo', 'utf-8')))
        Media seq=2 is out of sequence (expected 3) : FEC algorithm reseted !
        >>> print(g)
        Matrix size L x D            = 4 x 5
        Total invalid media packets  = 0
        Total media packets received = 5
        Column sequence number       = 1
        Row    sequence number       = 1
        Media  sequence number       = 3
        Medias buffer (seq. numbers) = [2]
        >>> if isinstance(g._medias[0].payload, bytearray):
        ...     assert g._medias[0].payload == bytearray('Kuota Kharma Evo', 'utf-8')
        ... else:
        ...     assert g._medias[0].payload == 'Kuota Kharma Evo'

        Testing a complete 3x4 matrix:

        >>> g = FecGenerator(3, 4)
        >>> R = RtpPacket
        >>> g.put_media(R.create(1, 100, R.MP2T_PT, bytearray('Tabby', 'utf-8')))
        Media seq=1 is out of sequence (expected None) : FEC algorithm reseted !
        >>> g.put_media(R.create(2, 200, R.MP2T_PT, bytearray('1234', 'utf-8')))
        >>> g.put_media(R.create(3, 300, R.MP2T_PT, bytearray('abcd', 'utf-8')))
        New ROW FEC packet seq=1 snbase=1 LxD=3xNone trec=384
        >>> g.put_media(R.create(4, 400, R.MP2T_PT, bytearray('python', 'utf-8')))
        >>> g.put_media(R.create(5, 500, R.MP2T_PT, bytearray('Kuota harma Evo', 'utf-8')))
        >>> g.put_media(R.create(6, 600, R.MP2T_PT, bytearray('h0ffman', 'utf-8')))
        New ROW FEC packet seq=2 snbase=4 LxD=3xNone trec=572
        >>> g.put_media(R.create(7, 700, R.MP2T_PT, bytearray('mutable', 'utf-8')))
        >>> g.put_media(R.create(8, 800, R.MP2T_PT, bytearray('10061987', 'utf-8')))
        >>> g.put_media(R.create(9, 900, R.MP2T_PT, bytearray('OSCIED', 'utf-8')))
        New ROW FEC packet seq=3 snbase=7 LxD=3xNone trec=536
        >>> g.put_media(R.create(10, 1000, R.MP2T_PT, bytearray('5ème élément', 'utf-8')))
        New COL FEC packet seq=1 snbase=1 LxD=3x4 trec=160
        >>> print(g)
        Matrix size L x D            = 3 x 4
        Total invalid media packets  = 0
        Total media packets received = 10
        Column sequence number       = 2
        Row    sequence number       = 4
        Media  sequence number       = 11
        Medias buffer (seq. numbers) = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        >>> g.put_media(R.create(11, 1100, R.MP2T_PT, bytearray('Chaos Theory', 'utf-8')))
        New COL FEC packet seq=2 snbase=2 LxD=3x4 trec=1616
        >>> g.put_media(R.create(12, 1200, R.MP2T_PT, bytearray('Yes, it WORKS !', 'utf-8')))
        New ROW FEC packet seq=4 snbase=10 LxD=3xNone trec=788
        New COL FEC packet seq=3 snbase=3 LxD=3x4 trec=1088
        >>> print(g)
        Matrix size L x D            = 3 x 4
        Total invalid media packets  = 0
        Total media packets received = 12
        Column sequence number       = 4
        Row    sequence number       = 5
        Media  sequence number       = 13
        Medias buffer (seq. numbers) = []
        """
        self._total += 1
        if not media.valid:
            self._invalid += 1
            return

        # Compute expected media sequence number for next packet
        sequence = (media.sequence + 1) & RtpPacket.S_MASK

        # Ensure that protected media packets are not out of sequence to generate valid FEC
        # packets. If media packet sequence number is not at attended value, it may mean :
        # - Looped VLC broadcast session restarted media
        # - Some media packet are really lost between the emitter and this software
        # - An unknown feature (aka bug) makes this beautiful tool crazy !
        if self._media_sequence is not None and media.sequence == self._media_sequence:
            self._medias.append(media)
        else:
            self._medias = [media]
            self.on_reset(media)
        self._media_sequence = sequence

        # Compute a new row FEC packet when a new row just filled with packets
        if len(self._medias) % self._L == 0:
            row_medias = self._medias[-self._L:]
            assert len(row_medias) == self._L
            row = FecPacket.compute(
                self._row_sequence,
                FecPacket.XOR,
                FecPacket.ROW,
                self._L,
                self._D,
                row_medias)
            self._row_sequence = (self._row_sequence + 1) & RtpPacket.S_MASK
            self.on_new_row(row)

        # Compute a new column FEC packet when a new column just filled with packets
        if len(self._medias) > self._L * (self._D - 1):
            first = len(self._medias) - self._L * (self._D - 1) - 1
            col_medias = self._medias[first::self._L]
            assert len(col_medias) == self._D
            col = FecPacket.compute(
                self._col_sequence,
                FecPacket.XOR,
                FecPacket.COL,
                self._L,
                self._D,
                col_medias)
            self._col_sequence = (self._col_sequence + 1) & RtpPacket.S_MASK
            self.on_new_col(col)

        if len(self._medias) == self._L * self._D:
            self._medias = []

    def __str__(self):
        """
        Returns a string containing a formated representation of the FEC streams generator.

        **Example usage**

        >>> print(FecGenerator(5, 6))
        Matrix size L x D            = 5 x 6
        Total invalid media packets  = 0
        Total media packets received = 0
        Column sequence number       = 1
        Row    sequence number       = 1
        Media  sequence number       = None
        Medias buffer (seq. numbers) = []
        """
        medias = [p.sequence for p in self._medias]
        return f"""Matrix size L x D            = {self._L} x {self._D}
Total invalid media packets  = {self._invalid}
Total media packets received = {self._total}
Column sequence number       = {self._col_sequence}
Row    sequence number       = {self._row_sequence}
Media  sequence number       = {self._media_sequence}
Medias buffer (seq. numbers) = {medias}"""



"""
Mix-ins for building your own `Django REST Framework <https://github.com/tomchristie/django-rest-framework>`_
powered API `metadata <https://github.com/tomchristie/django-rest-framework/blob/master/rest_framework/metadata.py>`_.
"""

from rest_framework import serializers

__all__ = ['ExcludeRelatedChoicesMixin']


class ExcludeRelatedChoicesMixin(object):
    """Do not includes related fields to avoid having choices with hundreds instances."""

    related_fields = (serializers.RelatedField, serializers.ManyRelatedField)

    def get_field_info(self, field):
        if hasattr(field, 'choices') and isinstance(field, self.related_fields):
            field_class = field.__class__

            class HaveNoChoicesProxy(field_class):

                @property
                def choices(self):
                    raise AttributeError

            try:
                field.__class__ = HaveNoChoicesProxy
                return super().get_field_info(field)
            finally:
                field.__class__ = field_class
        return super().get_field_info(field)



from pytoolbox import decorators
from . import brand, equipment  # pylint:disable=unused-import

__all__ = ['Lens']


class Lens(equipment.Equipement):

    brand_class = brand.Brand

    @property
    def brand(self):
        brands = {t.brand for t in self.tags.values() if t.brand}
        if brands:
            assert len(brands) == 1, brands
            return next(iter(brands))
        # Extract brand from model
        return self.brand_class(self._model.split(' ')[0])

    @property
    def _model(self):
        return next((t.data for t in self.tags.values() if 'model' in t.label.lower()), None)

    @decorators.cached_property
    def tags(self):
        return {k: t for k, t in self.metadata.tags.items() if 'lens' in t.label.lower()}



from pytoolbox.django.core.constants import DEFFERED_REGEX

__all__ = ['FieldsToValuesLookupDict']


class FieldsToValuesLookupDict(object):
    """
    Global registry for mapping X class fields to W values.

    * X can be a Model, a (Model)Form, a REST Framework Serializer, ...
    * Y can be the fields help texts or verbose names, or the number 42.

    Strange idea? Isn't it?

    Here is a short example as an appetizer. Suppose you want to define your application's help
    texts into a centralized registry, for keeping your wording DRY. And suppose you have some
    models like this:

    >> class Media(models.Model):
    ..     url = models.URLField()

    >> class File(models.Model):
    ..     url = models.URLField()

    And you instantiate this class with:

    >> help_texts = FieldsLookupDict({'Media.url': 'The media asset ingest URL', 'url': 'An URL'})

    Then, you can lookup for the help text of a field like this:

    >> help_texts[(Media, 'url')]
    The media asset ingest URL

    >> help_texts[(File, 'url')]
    An URL

    The value returned will be the first matching to the following keys:

    1. '<cls.__name__>.<field_name>'
    2. '<field_name>'

    If given class have a _meta or Meta ("meta") attribute with a model attribute, then the
    following keys are tried:

    1. '<cls.__name__>.<field_name>'
    2. '<cls._meta.model>.<field_name>'
    3. '<field_name>'
    """

    def __init__(self, name, translations=None):
        self.name = name
        self.translations = translations or {}

    def __getitem__(self, key):
        if isinstance(key, str):
            keys = [key]
        else:
            cls, field_name = key
            keys = [f'{cls.__name__}.{field_name}', field_name]
            meta = getattr(cls, '_meta', None) or getattr(cls, 'Meta', None)
            if meta and hasattr(meta, 'model'):
                # cleanup model name when some fields are deferred (Media vs Media_Deffered_...)
                keys.insert(1, f"{DEFFERED_REGEX.sub('', meta.model.__name__)}.{field_name}")
        for key in keys:
            value = self.translations.get(key)
            if value:
                return value
        raise KeyError(f'Entry for keys {keys} not found in {self.name}.')

    def __setitem__(self, key, value):
        self.translations[key] = value



import atexit, code, os, sys

__all__ = ['confirm', 'choice', 'print_error', 'progress_bar', 'shell']


def confirm(question=None, default=False, stream=sys.stdout):
    """
    Return True if user confirm the action, else False. `default` if user only press ENTER.

    **Example usage**

    ::

        >> confirm('Do it now', default=True)
        Do it now ? [Y/n]:
        True
        >> confirm('Are you sure', default=False)
        Are you sure ? [y/N]:
        False
        >> confirm('Really, I am sure that is false', default=False)
        Really, I am sure that is false ? [y/N]: y
        True
    """
    if question is None:
        question = 'Confirm'
    question = f"{question} ? [{'Y/n' if default else 'y/N'}]: "
    while True:
        stream.write(question)
        stream.flush()
        answer = input()
        if not answer:
            return default
        if answer.lower() in ('y', 'yes'):
            return True
        if answer.lower() in ('n', 'no'):
            return False
        stream.write(f'please enter y(es) or n(o).{os.linesep}')
        stream.flush()


def choice(question='', choices=tuple(), stream=sys.stdout):
    """
    Prompt the user for a choice and return his/her answer.

    **Example of usage**

    ::

        >> choice('What is your favorite color?', ['blue', 'orange', 'red'])
        What is your favourite color? [blue, orange, red]: orange
        orange
        >> choice(['male', 'female'])
        [male, female]? female
        female
    """

    # generate question and choices list
    choices_string = ', '.join(choices)
    if question is None:
        question = f'[{choices_string}]? '
    else:
        question = f'{question} [{choices_string}]: '

    # loop until an acceptable choice has been answered
    while True:
        ans = input(question)
        if ans in choices:
            return ans
        stream.write(f'Please choose between {choices_string}.{os.linesep}')


def print_error(message, exit_code=1, stream=sys.stderr):
    """
    Print an error message and exit if `exit_code` is not None.

    **Example usage**

    In following example stream is set to `sys.stdout` and exit is disabled (for :mod:`doctest`):

    >>> print_error("It's not a bug - it's an undoc. feature.", exit_code=None, stream=sys.stdout)
    [ERROR] It's not a bug - it's an undoc. feature.
    """
    stream.write(f'[ERROR] {message}{os.linesep}')
    stream.flush()
    if exit_code is not None:
        sys.exit(exit_code)


def progress_bar(
    start_time,  # pylint:disable=unused-argument
    current,
    total,
    size=50,
    done='=',
    todo=' ',
    template='\r[{done}{todo}]',
    stream=sys.stdout
):
    """
    Show a progress bar. Default `template` string starts with a carriage return to update progress
    on same line.

    **Example usage**

    >>> import functools, time
    >>> progress = functools.partial(progress_bar, template='[{done}{todo}]', stream=sys.stdout)
    >>> progress(time.time(), 10, 15, size=30)
    [====================          ]
    >>> progress(time.time(), 1, 6, size=10)
    [=         ]
    >>> progress(time.time(), 3, 5, size=5, done='+', todo='-')
    [+++--]
    """
    if total:
        progress = int(size * current / total)
        stream.write(template.format(done=done * progress, todo=todo * (size - progress)))
        stream.flush()


def shell(banner=None, history_filename='~/.python_history', history_length=1000, imported=None):
    """Execute an interactive shell with auto-completion and history (if available)."""
    # Setup auto-completion and file history, credits to @kyouko-taiga!
    try:
        import readline
    except ImportError:
        pass
    else:
        import rlcompleter
        readline.set_completer(rlcompleter.Completer(imported).complete)
        readline.parse_and_bind("tab:complete")
        history_filename = os.path.abspath(os.path.expanduser(history_filename))
        try:
            readline.read_history_file(history_filename)
            readline.set_history_length(history_length)
        except FileNotFoundError:
            pass
        atexit.register(readline.write_history_file, history_filename)
    return code.interact(banner=banner, local=imported)


if __name__ == '__main__':

    if confirm('Please confirm this'):
        print('You confirmed')
    else:
        print('You do not like my question')

    print(choice('Select a language', ['Italian', 'French']))



import inspect, io, traceback

from . import module

_all = module.All(globals())


class MessageMixin(Exception):

    message = None

    def __init__(self, message=None, **kwargs):
        if message is not None:
            self.message = message
        self.__dict__.update(kwargs)
        Exception.__init__(self)

    def __str__(self):
        attributes = inspect.getmembers(self, lambda a: not inspect.isroutine(a))
        return self.message.format(**{a: v for a, v in attributes if a[0] != '_'})


class BadHTTPResponseCodeError(MessageMixin, Exception):
    message = 'Download request {url} code {r_code} expected {code}.'


class CorruptedFileError(MessageMixin, Exception):
    message = 'File {path} is corrupted checksum {file_hash} expected {expected_hash}.'


class ForbiddenError(Exception):
    """A forbidden error."""


class InvalidBrandError(MessageMixin, Exception):
    message = 'Brand {brand} not in {brands}.'


class InvalidIPSocketError(MessageMixin, Exception):
    message = '{socket} is not a valid IP socket.'


class MultipleSignalHandlersError(MessageMixin, Exception):
    message = 'Signal {signum} already handled by {handlers}.'


class UndefinedPathError(Exception):
    pass


def assert_raises_item(exception_cls, something, index, value=None, delete=False):
    """

    **Example usage**

    >>> x = {0: 3.14, 1: 2.54}

    Assert that __getitem__ will fail:

    >>> assert_raises_item(KeyError, x, 2)
    >>> assert_raises_item(ValueError, x, 3)
    Traceback (most recent call last):
        ...
    ValueError: Exception KeyError is not an instance of ValueError.
    >>> assert_raises_item(Exception, x, 0)
    Traceback (most recent call last):
        ...
    AssertionError: Exception Exception not raised.

    Assert that __setitem__ will fail:

    >>> assert_raises_item(TypeError, x, [10], value=3.1415)
    >>> assert_raises_item(TypeError, x, 0, value=3.1415)
    Traceback (most recent call last):
        ...
    AssertionError: Exception TypeError not raised.

    Assert that __delitem__ will fail:

    >>> assert_raises_item(KeyError, x, 2, delete=True)
    >>> assert_raises_item(KeyError, x, 1, delete=True)
    Traceback (most recent call last):
        ...
    AssertionError: Exception KeyError not raised.

    >>> x == {0: 3.1415}
    True
    """
    try:
        if delete:
            del something[index]
        elif value is None:
            something[index]  # pylint:disable=pointless-statement
        else:
            something[index] = value
    except Exception as e:  # pylint:disable=broad-except
        if not isinstance(e, exception_cls):
            raise ValueError(
                f'Exception {e.__class__.__name__} is not '
                f'an instance of {exception_cls.__name__}.') from e
        return
    raise AssertionError(f'Exception {exception_cls.__name__} not raised.')


def get_exception_with_traceback(exception):
    """
    Return a string with the exception traceback.

    **Example usage**

    If the exception was not raised then there are no traceback:

    >>> get_exception_with_traceback(ValueError('yé'))
    'ValueError: yé\\n'

    If the exception was raised then there is a traceback:

    >>> try:
    ...     raise RuntimeError('yé')
    ... except Exception as e:
    ...     trace = get_exception_with_traceback(e)
    >>> 'Traceback' in trace
    True
    >>> "raise RuntimeError('yé')" in trace
    True
    """
    buf = io.StringIO()
    traceback.print_exception(type(exception), exception, exception.__traceback__, file=buf)
    return buf.getvalue()


__all__ = _all.diff(globals())



"""
Mix-ins for building your own models fields.
"""

from pytoolbox.django.core import validators

__all__ = ['LowerCaseMixin', 'OptionsMixin', 'StripMixin']


class LowerCaseMixin(object):

    def get_prep_value(self, value):
        value = super().get_prep_value(value)
        if value is not None:
            value = value.lower()
        return value


class NullifyMixin(object):
    """Replace falsy values by None if NULL is allowed."""

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)
        if not value and self.null:
            value = None
            setattr(model_instance, self.attname, value)
        return value


class OptionsMixin(object):

    default_options = {}
    override_options = {}

    def __init__(self, **kwargs):
        super().__init__(**{**self.default_options, **kwargs, **self.override_options})


class StripMixin(object):
    """https://code.djangoproject.com/ticket/6362#no1"""

    default_validators = [validators.EmptyValidator()]

    def pre_save(self, model_instance, add):
        value = super().pre_save(model_instance, add)
        if value:
            value = value.strip()
            setattr(model_instance, self.attname, value)
        return value



"""
Mix-ins for building your own models.

Recommended sub-classing order:

- BetterUniquenessErrorsMixin
- AutoForceInsertMixin
- CallFieldsPreSaveMixin
- AutoUpdateFieldsMixin
- AlwaysUpdateFieldsMixin
- AutoRemovePKFromUpdateFieldsMixin
- ValidateOnSaveMixin (will defeat the detection method of AutoUpdateFieldsMixin if placed before it)
- UpdatePreconditionsMixin
- StateTransitionPreconditionMixin
- StateTransitionEventsMixin

Order for these does not matter:

- PublicMetaMixin
- RelatedModelMixin
- ReloadMixin
- SaveInstanceFilesMixin
"""

import collections, itertools, re

from django.core.exceptions import ValidationError
from django.db import DatabaseError
from django.db.models.fields.files import FileField
from django.db.utils import IntegrityError
from django.utils.functional import cached_property

from pytoolbox.django import signals
from pytoolbox.django.core import exceptions
from pytoolbox import itertools as py_itertools, module  # pylint:disable=reimported
from . import utils

_all = module.All(globals())


class AlwaysUpdateFieldsMixin(object):
    """
    Ensure fields listed in the attribute ``self.always_update_fields`` are always updated by
    ``self.save()``. Makes the usage of ``self.save(update_fields=...)`` cleaner.
    """
    def save(self, *args, **kwargs):
        update_fields = kwargs.get('update_fields')
        if kwargs.get('force_update') or update_fields:
            update_fields = set(update_fields or [])
            update_fields.update(self.always_update_fields)
            kwargs['update_fields'] = update_fields
        super().save(*args, **kwargs)


class AutoForceInsertMixin(object):

    def save(self, *args, **kwargs):
        if kwargs.get('force_insert') is None:
            kwargs['force_insert'] = self._state.adding
        super().save(*args, **kwargs)


class AutoRemovePKFromUpdateFieldsMixin(object):
    """
    If the primary key is set but unchanged, then remove the primary key from the list of fields to
    update. This fix an issue when saving, ``ValueError: The following fields do not exist in this
    model or are m2m fields: id.``. This check is probably implemented by Django developers to
    protect from unintentional data overwrite.

    If the primary key is set to a new value, then the intention of the developer is probably to
    duplicate the model by saving it with a new primary key. So the mix-in let Django save with its
    own default options for save.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.previous_pk = self.pk

    def save(self, **kwargs):
        update_fields = kwargs.get('update_fields')
        if update_fields and self._meta.pk.attname in update_fields:
            if self.pk == self.previous_pk:
                kwargs['update_fields'] = {f for f in update_fields if f != self._meta.pk.attname}
            else:
                # This is probably the model duplication pattern
                for argument in 'force_insert', 'force_update', 'update_fields':
                    kwargs.pop(argument, None)
        super().save(**kwargs)
        self.previous_pk = self.pk


class AutoUpdateFieldsMixin(object):
    """
    Keep track of what fields were set in order to make UPDATE queries lighter.

    This mix-in comes with the following features:

    * Foreign keys and the mutable types are correctly handled.
    * Models with a primary key preset to a value before being saved in database are correctly
      handled.
    * You can specify the value for `force_update` if it is `None` with `default_force_update`.

    However this low-memory footprint mix-in also comes with some limitations, it does not:

    * Store old fields values - you cannot know if the fields are really modified or not.
    * Watch for background modifications of the mutable fields - it can drives you crazy, sometimes.
    * Detect fields updated by the field's pre_save - CallFieldsPreSaveMixin before this mix-in.
    * Filter the primary key from the list of fields to update - AutoRemovePKFromUpdateFieldsMixin
      after this mix-in.
    """
    default_force_update = False

    def __init__(self, *args, **kwargs):
        super(AutoUpdateFieldsMixin, self).__init__(*args, **kwargs)
        self._setted_fields = set()
        self._fields_names = frozenset(f.attname for f in self._meta.fields)

    def __setattr__(self, name, value):
        # Check the instance is both initialized and already stored in DB
        if name in getattr(self, '_fields_names', set()) and not self._state.adding:
            self._setted_fields.add(name)
        return super().__setattr__(name, value)

    def save(self, *args, **kwargs):
        if not self._state.adding and not kwargs.get('force_insert'):
            if kwargs.get('force_update') is None:
                kwargs['force_update'] = self.default_force_update
            if kwargs.get('update_fields') is None:
                kwargs['update_fields'] = set(self._setted_fields)
        super().save(*args, **kwargs)
        self._setted_fields = set()


class BetterUniquenessErrorsMixin(object):
    """
    Hide some fields from the unique-together errors.
    Convert uniqueness integrity errors to validation errors.

    **Use cases**

    Case: Your model have some non editable fields that are included in a uniqueness constraint.
    Issue: The uniqueness errors shown in your forms includes the name of the hidden fields.
    Solution: Exclude the name of the hidden fields from the error messages.
    Implementation: Set `unique_together_hide_fields` to the name of those hidden fields.

    Case: Sometimes the form submit may raise an integrity error (concurrency, [yes] crazy
          unexpected usage of Django).
    Issue: Those integrity errors are unfortunately returned as Internal Server Error (HTTP 500).
    Solution: Convert the uniqueness integrity errors to validation errors and return an awesome
              form with errors.
    Implementation: Set `unique_from_integrity_error` to True. And subclass
    :class:`pytoolbox.django.views.mixins.ValidationErrorsMixin` in your edit views.
    """
    unique_from_integrity_error = True
    unique_together_hide_fields = ()

    def save(self, *args, **kwargs):
        try:
            super().save(*args, **kwargs)
        except IntegrityError as e:
            if self.unique_from_integrity_error:
                match = re.search(r'duplicate key[^\)]+\((?P<fields>[^\)]+)\)', e.args[0])
                if match:
                    fields = {
                        f.strip().replace('_id', '')
                        for f in match.groupdict()['fields'].split(',')
                    }
                    if fields in (set(u) for u in self._meta.unique_together):
                        fields = sorted(fields - set(self.unique_together_hide_fields))
                        if fields:
                            error = self.unique_error_message(self.__class__, fields)
                            raise ValidationError({fields[0]: error}) if len(fields) == 1 else error
                        return self._handle_hidden_duplicate_key_error(e)
            raise

    def _handle_hidden_duplicate_key_error(self, e):
        raise e

    def _perform_unique_checks(self, unique_checks):
        errors_by_field = super()._perform_unique_checks(unique_checks)
        hidden_fields = set(self.unique_together_hide_fields)
        if not hidden_fields:
            return errors_by_field
        filtered_errors_by_field = collections.defaultdict(list)
        for field, errors in errors_by_field.items():
            for error in errors:
                # only process the uniqueness errors related to multiple fields
                if len(error.params.get('unique_check', [])) > 1:
                    fields = [f for f in error.params['unique_check'] if f not in hidden_fields]
                    if fields:
                        error = self.unique_error_message(self.__class__, fields)
                        filtered_errors_by_field[
                            fields[0] if len(fields) == 1 else field
                        ].append(error)
                else:
                    filtered_errors_by_field[field].append(error)
        return filtered_errors_by_field


class CallFieldsPreSaveMixin(object):
    """
    If you wanna be sure the fields pre_save method are called, now you can!

    For more information see: https://code.djangoproject.com/ticket/25363
    """
    def save(self, *args, **kwargs):
        non_pk_fields = (f for f in self._meta.local_concrete_fields if not f.primary_key)
        for field in non_pk_fields:
            field.pre_save(self, self._state.adding)
        super().save(*args, **kwargs)


class PublicMetaMixin(object):
    """
    Make `_meta` public in templates through a class method called `meta`.
    """
    @classmethod
    def meta(cls):
        return cls._meta


class RelatedModelMixin(object):

    @classmethod
    def get_related_manager(cls, field):
        return utils.get_related_manager(cls, field)

    @classmethod
    def get_related_model(cls, field):
        return utils.get_related_model(cls, field)


class ReloadMixin(object):

    def reload(self):
        return self._meta.model._default_manager.get(pk=self.pk)


class SaveInstanceFilesMixin(object):
    """
    Overrides saves() with a method that saves the instance first and then the instance's file
    fields this ensure that the upload_path method will get a valid instance id / private key.
    """
    def save(self, *args, **kwargs):
        saved_fields = {}
        if self.pk is None:
            for field in self._meta.fields:
                if isinstance(field, FileField):
                    saved_fields[field.name] = getattr(self, field.name)
                    setattr(self, field.name, None)
            super().save(*args, **kwargs)
            for name, value in saved_fields.items():
                setattr(self, name, value)
            kwargs['force_insert'] = False  # Do not force because we already saved the instance
        super().save(*args, **kwargs)


class UpdatePreconditionsMixin(object):

    precondition_error_class = exceptions.DatabaseUpdatePreconditionsError

    def apply_preconditions(self, base_qs, using, pk_val, values, update_fields, force_update):
        if hasattr(self, '_preconditions'):
            pre_excludes, pre_filters = self._preconditions
            del self._preconditions
            if pre_excludes:
                base_qs = base_qs.exclude(**pre_excludes)
            if pre_filters:
                base_qs = base_qs.filter(**pre_filters)
        return base_qs, using, pk_val, values, update_fields, force_update

    def pop_preconditions(self, *args, **kwargs):
        self._preconditions = kwargs.pop('pre_excludes', {}), kwargs.pop('pre_filters', {})
        return args, kwargs, any(self._preconditions)

    def save(self, *args, **kwargs):
        args, kwargs, has_preconditions = self.pop_preconditions(*args, **kwargs)
        try:
            super(UpdatePreconditionsMixin, self).save(*args, **kwargs)
        except DatabaseError as e:
            if has_preconditions and 'did not affect' in '{0}'.format(e):
                raise self.precondition_error_class()
            raise

    def _do_update(self, base_qs, using, pk_val, values, update_fields, force_update):
        # FIXME _do_update is called once for each model in the inheritance hierarchy: Handle this!
        args = self.apply_preconditions(base_qs, using, pk_val, values, update_fields, force_update)
        updated = super()._do_update(*args)
        if not updated and args[0] != base_qs and base_qs.filter(pk=pk_val).exists():
            raise self.precondition_error_class()
        return updated


class StateTransitionEventsMixin(object):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.previous_state = self.state

    def on_post_state_transition(self, args, kwargs):
        signals.post_state_transition.send(
            instance=self, previous_state=self.previous_state, args=args, kwargs=kwargs)

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if 'state' in kwargs.get('update_fields', ['state']):
            self.on_post_state_transition(args, kwargs)
            self.previous_state = self.state


class StateTransitionPreconditionMixin(UpdatePreconditionsMixin):

    check_state = True
    invalid_state_error_class = exceptions.InvalidStateError
    transition_not_allowed_error_class = exceptions.TransitionNotAllowedError

    def can_transit_to(self, state, fail=False, noop_skip=False):
        """
        Helper that return the following:

        * True if transition to `state` is allowed.
        * False if `fail` is set to False or `noop_skip` is set to True and state is unchanged.
        * Else raise a transition not allowed error.
        """
        if state in self.states.TRANSITIONS[self.state]:
            return True
        if not fail or noop_skip and self.state == state:
            return False
        raise self.transition_not_allowed_error_class(instance=self, state=state)

    def check_state_in(self, states, fail=False):
        states = sorted(py_itertools.chain(states))
        if self.state in states:
            return True
        if not fail:
            return False
        raise self.invalid_state_error_class(instance=self, states=states)

    def pop_preconditions(self, *args, **kwargs):
        """
        Add state precondition if state will be saved and state is not enforced by preconditions.
        """
        args, kwargs, has_preconditions = super().pop_preconditions(*args, **kwargs)
        if (
            kwargs.pop('check_state', self.check_state) and
            'state' in kwargs.get('update_fields', ['state'])
        ):
            pre_excludes, pre_filters = self._preconditions
            if not any(f.startswith('state') for f in itertools.chain(pre_excludes, pre_filters)):
                states, valid = self.states.get_transit_from(self.state, auto_inverse=True)
                assert states, (states, valid)
                key, values = (
                    ('state__in', states) if len(states) > 1 else ('state', next(iter(states)))
                )
                (pre_filters if valid else pre_excludes)[key] = values
        return args, kwargs, any(self._preconditions)


class ValidateOnSaveMixin(object):

    validate_on_save = True
    validate_on_save_kwargs = {}

    def save(self, *args, **kwargs):
        if kwargs.pop('validate', self.validate_on_save):
            self.full_clean(**self.validate_on_save_kwargs)
        super().save(*args, **kwargs)


class FasterValidateOnSaveMixin(ValidateOnSaveMixin):
    """
    Do not validate uniqueness nor relation fields on save to prevent excessive SELECT queries.
    """

    @cached_property
    def validate_on_save_kwargs(self):
        return {
            'exclude': [f.name for f in self._meta.concrete_fields if f.rel],
            'validate_unique': False
        }


__all__ = _all.diff(globals())



"""
Mix-ins for building your own `Django REST Framework <https://github.com/tomchristie/django-rest-framework>`_
powered API `views <http://www.django-rest-framework.org/tutorial/3-class-based-views/>`_.
"""

from django.contrib.auth.views import redirect_to_login
from rest_framework import renderers

from pytoolbox import module

_all = module.All(globals())


class ActionToQuerysetMixin(object):

    querysets = {}

    def get_queryset(self):
        return self.querysets.get(self.action, self.queryset)


class ActionToSerializerMixin(object):

    serializers_classes = {}

    def get_serializer_class(self):
        return self.serializers_classes.get(self.action, self.serializer_class)


class MethodToQuerysetMixin(object):

    querysets = {}

    def get_queryset(self):
        return self.querysets.get(self.request.method, self.queryset)


class MethodToSerializerMixin(object):

    serializers_classes = {}

    def get_serializer_class(self):
        return self.serializers_classes.get(self.request.method, self.serializer_class)


class RedirectToLoginMixin(object):

    redirected_classes = (renderers.BrowsableAPIRenderer, )

    def finalize_response(self, request, response, *args, **kwargs):
        response = super().finalize_response(request, response, *args, **kwargs)
        logged = request.user.is_authenticated
        if (
            not (logged if isinstance(logged, bool) else logged()) and
            isinstance(response.accepted_renderer, self.redirected_classes)
        ):
            response = redirect_to_login(request.path)
            response.data = {}
        return response


__all__ = _all.diff(globals())



import configparser, os, re

__all__ = ['CONFIG_PREFIX', 'DRIVER_IN_KERNEL', 'DRIVER_HAS_MODULE', 'get_kernel_config']

CONFIG_PREFIX = re.compile(r'^config_')
DRIVER_IN_KERNEL = 'y'
DRIVER_HAS_MODULE = 'm'


def get_kernel_config(release=None, fail=True):
    """
    Return a JSON string with the GNU/Linux Kernel configuration.

    **Example usage**

    >>> config = get_kernel_config(fail=False)
    >>> type(config)
    <class 'dict'>
    >>> not config or 'memory' in config
    True

    Error handling:

    >>> get_kernel_config('0.0.1-generic', fail=False)
    {}
    """
    try:
        with open(f'/boot/config-{release or os.uname().release}', encoding='utf-8') as f:
            config = configparser.ConfigParser()
            config.read_string(f'[kernel]{f.read()}')
    except IOError:
        if fail:
            raise
        return {}
    return {CONFIG_PREFIX.sub('', k): v for k, v in config.items('kernel')}



import datetime, numbers
from calendar import timegm
from time import mktime

import pytz

from . import module

_all = module.All(globals())


def datetime_now(
    format='%Y-%m-%d %H:%M:%S',
    append_utc=False,
    offset=None,
    tz=pytz.utc
):  # pylint:disable=redefined-builtin
    """
    Return the current (timezone aware) date and time as UTC, local (tz=None) or related to a
    timezone. If `format` is not None, the date will be returned in a formatted string.

    :param format: Output date string formatting
    :type format: str
    :param append_utc: Append ' UTC' to date string
    :type append_utc: bool
    :param offset: Offset to add to current time
    :type offset: datetime.timedelta
    :param tz: The timezone (e.g. ``pytz.timezone('EST')``)
    :type tz: tz

    **Example usage**

    Add an offset:

    >>> now = datetime_now(format=None)
    >>> future = datetime_now(offset=datetime.timedelta(hours=2, minutes=10), format=None)
    >>> result = (future - now)
    >>> type(result)
    <class 'datetime.timedelta'>
    >>> print(result)
    2:10:00.00...

    Append UTC to output date string:

    >>> type(datetime_now())
    <class 'str'>
    >>> assert ' UTC' not in datetime_now(tz=pytz.utc, append_utc=False)
    >>> assert ' UTC' not in datetime_now(tz=None, append_utc=True)
    >>> assert ' UTC' not in datetime_now(tz=pytz.timezone('EST'), append_utc=True)
    >>> assert ' UTC' in datetime_now(tz=pytz.utc, append_utc=True)

    Play with timezones::

        >> datetime_now(tz=pytz.timezone('Europe/Zurich'))
        '2013-10-17 09:54:08'
        >> datetime_now(tz=pytz.timezone('US/Eastern'))
        '2013-10-17 03:54:08'
    """
    now = datetime.datetime.now(tz)
    if offset is not None:
        now += offset
    if not format:
        return now
    return now.strftime(format) + (' UTC' if tz == pytz.utc and append_utc else '')


def datetime_to_str(
    date_time,
    format='%Y-%m-%d %H:%M:%S',
    append_utc=False
):  # pylint:disable=redefined-builtin
    return date_time.strftime(format) + (' UTC' if append_utc else '')


def str_to_datetime(
    date,
    format='%Y-%m-%d %H:%M:%S',
    fail=True
):  # pylint:disable=redefined-builtin
    """
    Return the `date` string converted into an instance of :class:`datetime.datetime`.
    Handle 24h+ hour format like 2015:06:28 24:05:00 equal to the 28th June 2015 at
    midnight and 5 minutes.

    **Example usage**

    >>> str_to_datetime('1985-01-06 05:02:00')
    datetime.datetime(1985, 1, 6, 5, 2)
    >>> str_to_datetime('this is not a date')
    Traceback (most recent call last):
        ...
    ValueError: time data 'this is not a date' does not match format '%Y-%m-%d %H:%M:%S'
    """
    try:
        return datetime.datetime.strptime(date.replace(': ', ':0').replace(' 24:', ' 00:'), format)
    except ValueError:
        if fail and date != '0000:00:00 00:00:00':
            raise
    return None


def multiply_time(value, factor, as_delta=False):
    """
    Return an instance of :class:`datetime.time`/:class:`datetime.timedelta`
    corresponding to `value` multiplied by a `factor`.

    **Example usage**

    >>> multiply_time('00:10:00', 0.5)
    datetime.time(0, 5)
    >>> multiply_time(datetime.timedelta(seconds=60), 3)
    datetime.time(0, 3)
    >>> multiply_time(120, 0.1)
    datetime.time(0, 0, 12)
    >>> res = multiply_time(datetime.timedelta(seconds=152, microseconds=500000), 1, as_delta=True)
    >>> type(res)
    <class 'datetime.timedelta'>
    >>> print(res)
    0:02:32.500000
    """
    return secs_to_time(total_seconds(value) * factor, as_delta=as_delta)


def parts_to_time(hours, minutes, seconds, microseconds, as_delta=False):
    """
    Return an instance of :class:`datetime.time`/:class:`datetime.timedelta` out of the parts.

    **Example usage**

    >>> parts_to_time(23, 15, 7, 3500)
    datetime.time(23, 15, 7, 3500)
    >>> result = parts_to_time(23, 15, 7, 3500, as_delta=True)
    >>> type(result)
    <class 'datetime.timedelta'>
    >>> print(result)
    23:15:07.003500
    """
    if as_delta:
        return datetime.timedelta(
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            microseconds=microseconds)
    return datetime.time(hours, minutes, seconds, microseconds)


def secs_to_time(value, defaults_to_zero=False, as_delta=False):
    """
    Return an instance of :class:`datetime.time`/:class:`datetime.timedelta`, taking `value` as the
    number of seconds + microseconds (e.g. 10.3 = 10s 3000us).

    **Example usage**

    >>> secs_to_time(83707.0035)
    datetime.time(23, 15, 7, 3500)
    >>> secs_to_time(None)
    >>> secs_to_time(None, defaults_to_zero=True)
    datetime.time(0, 0)
    >>> result = secs_to_time(83707.0035, as_delta=True)
    >>> type(result)
    <class 'datetime.timedelta'>
    >>> print(result)
    23:15:07.003500
    >>> secs_to_time(None, as_delta=True) is None
    True
    >>> secs_to_time(None, defaults_to_zero=True, as_delta=True)
    datetime.timedelta(0)
    """
    try:
        delta = datetime.timedelta(seconds=float(value))
        return delta if as_delta else (datetime.datetime.min + delta).time()
    except (ValueError, TypeError):
        if defaults_to_zero and not value:
            return datetime.timedelta(seconds=0) if as_delta else datetime.time(second=0)
        return None


def str_to_time(value, defaults_to_zero=False, as_delta=False):
    """
    Return the string of format 'hh:mm:ss' into an instance of time.

    **Example usage**

    >>> str_to_time('08:23:57')
    datetime.time(8, 23, 57)
    >>> str_to_time('00:03:02.12')
    datetime.time(0, 3, 2, 120)
    >>> str_to_time(None)
    >>> str_to_time(None, defaults_to_zero=True)
    datetime.time(0, 0)
    >>> result = str_to_time('08:23:57', as_delta=True)
    >>> type(result)
    <class 'datetime.timedelta'>
    >>> str(result)
    '8:23:57'
    >>> result = str_to_time('00:03:02.12', as_delta=True)
    >>> type(result)
    <class 'datetime.timedelta'>
    >>> str(result)
    '0:03:02.120000'
    >>> str_to_time(None, as_delta=True) is None
    True
    >>> str_to_time(None, defaults_to_zero=True, as_delta=True)
    datetime.timedelta(0)
    """
    try:
        hours, minutes, seconds_float = value.split(':')
        hours, minutes, seconds_float = int(hours), int(minutes), float(seconds_float)
        if as_delta:
            return datetime.timedelta(hours=hours, minutes=minutes, seconds=seconds_float)
        seconds = int(seconds_float)
        return datetime.time(hours, minutes, seconds, int(1000 * (seconds_float - seconds)))
    except (AttributeError, TypeError, ValueError):
        if defaults_to_zero and not value:
            return datetime.timedelta(seconds=0) if as_delta else datetime.time(second=0)
        return None


def time_ratio(numerator, denominator, zero_div_result=1.0):
    """
    Return the ratio between two times.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> time_ratio('0:30:00', '01:30:00')
    0.33...
    >>> time_ratio('0:00:05', '00:00:00')
    1.0
    >>> with asserts.raises(ValueError):
    ...     time_ratio('01:42:34', 'N/A')
    """
    try:
        ratio = total_seconds(numerator) / total_seconds(denominator)
        return 0.0 if ratio < 0.0 else 1.0 if ratio > 1.0 else ratio
    except ZeroDivisionError:
        return zero_div_result


def total_seconds(time):
    """
    Return the `time` converted in seconds.

    **Example usage**

    >>> total_seconds('00:10:00')
    600.0
    >>> total_seconds('01:54:17')
    6857.0
    >>> round(total_seconds('16.40'), 3)
    16.4
    >>> total_seconds(143.2)
    143.2
    >>> total_seconds(datetime.timedelta(seconds=152, microseconds=500000))
    152.5
    >>> total_seconds(datetime.datetime(2010, 6, 10, 0, 1, 30))
    90.0
    >>> total_seconds(datetime.datetime(2010, 6, 10, 14, 15, 23))
    51323.0
    >>> total_seconds(datetime.datetime(2010, 6, 10, 23, 59, 59))
    86399.0
    """
    try:
        if isinstance(time, datetime.timedelta):
            return time.total_seconds()
        if isinstance(time, numbers.Number):
            return time
        if isinstance(time, str):
            hours, minutes, seconds = time.split(':')
        else:
            hours, minutes, seconds = time.hour, time.minute, time.second
        return int(hours) * 3600 + int(minutes) * 60 + float(seconds)
    except ValueError:
        return float(time)


def datetime_to_epoch(date_time, utc=True, factor=1):
    """
    Return the :class:`datetime.datetime`/:class:`datetime.date` converted into an Unix epoch.
    Default `factor` means that the result is in seconds.

    **Example usage**

    >>> datetime_to_epoch(datetime.datetime(1970, 1, 1), factor=1)
    0
    >>> datetime_to_epoch(datetime.datetime(2010, 6, 10))
    1276128000
    >>> datetime_to_epoch(datetime.datetime(2010, 6, 10), factor=1000)
    1276128000000
    >>> datetime_to_epoch(datetime.date(2010, 6, 10), factor=1000)
    1276128000000
    >>> datetime_to_epoch(datetime.date(1987, 6, 10), factor=1000)
    550281600000

    In Switzerland::

        >> datetime_to_epoch(datetime.datetime(1970, 1, 1), utc=False, factor=1)
        -3600
        >> datetime_to_epoch(datetime.date(1970, 1, 1), utc=False, factor=1)
        -3600
    """
    if utc:
        has_it = hasattr(date_time, 'utctimetuple')
        time_tuple = date_time.utctimetuple() if has_it else date_time.timetuple()
        return int(timegm(time_tuple) * factor)
    return int(mktime(date_time.timetuple()) * factor)


def epoch_to_datetime(unix_epoch, tz=pytz.utc, factor=1):
    """
    Return the Unix epoch converted to a :class:`datetime.datetime`.
    Default `factor` means that the `unix_epoch` is in seconds.

    **Example usage**

    >>> epoch_to_datetime(0, factor=1)
    datetime.datetime(1970, 1, 1, 0, 0, tzinfo=<UTC>)
    >>> epoch_to_datetime(1276128000, factor=1)
    datetime.datetime(2010, 6, 10, 0, 0, tzinfo=<UTC>)
    >>> epoch_to_datetime(1276128000, tz=pytz.timezone('Europe/Zurich'), factor=1)
    datetime.datetime(2010, 6, 10, 2, 0, tzinfo=<DstTzInfo 'Europe/Zurich' CEST+2:00:00 DST>)
    >>> epoch_to_datetime(1276128000000, factor=1000)
    datetime.datetime(2010, 6, 10, 0, 0, tzinfo=<UTC>)
    >>> today = datetime.datetime(1985, 6, 1, 5, 2, 0, tzinfo=pytz.utc)
    >>> epoch_to_datetime(datetime_to_epoch(today, factor=1000), factor=1000) == today
    True
    """
    return datetime.datetime.fromtimestamp(unix_epoch / factor, tz)


__all__ = _all.diff(globals())



import collections, hashlib, os, random, string

from . import filesystem

__all__ = ['checksum', 'new', 'get_password_generator', 'githash', 'guess_algorithm']


def new(algorithm=hashlib.sha256):
    """
    Return an instance of a hash algorithm from :mod:`hashlib` if `algorithm`
    is a string else instantiate algorithm.
    """
    return hashlib.new(algorithm) if isinstance(algorithm, str) else algorithm()


def checksum(
    path_or_data,
    encoding='utf-8',
    is_path=False,
    algorithm=hashlib.sha256,
    chunk_size=None
):
    """
    Return the result of hashing `data` by given hash `algorithm`.

    **Example usage**

    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> checksum('')
    'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    >>> checksum('', algorithm=hashlib.md5)
    'd41d8cd98f00b204e9800998ecf8427e'
    >>> checksum('give me some hash please')
    'cebf462dd7771c78d3957446b1b4a2f5928731ca41eff66aa8817a6513ea1313'
    >>> checksum('et ça fonctionne !\\n')
    'ced3a2b067d105accb9f54c0b37eb79c9ec009a61fee5df7faa8aefdbff1ddef'
    >>> checksum('et ça fonctionne !\\n', algorithm='md5')
    '3ca34e7965fd59beaa13b6e7094f43e7'
    >>> checksum(directory / '..' / 'setup.py', is_path=True)
    '28ffad590e8d5cc888a6fcb77cab07252692d29802206ba81945afb69aab358d'
    >>> checksum(directory / '..' / 'setup.py', is_path=True, chunk_size=997)
    '28ffad590e8d5cc888a6fcb77cab07252692d29802206ba81945afb69aab358d'
    """
    hasher = new(algorithm)
    for data in filesystem.get_bytes(path_or_data, encoding, is_path, chunk_size):
        hasher.update(data)
    return hasher.hexdigest()


def get_password_generator(characters=string.ascii_letters + string.digits, length=16):
    """
    Return a dead simple password generator in the form of a dictionary with
    missing keys generated on the fly.

    **Example usage**

    >>> import re
    >>> passwords = get_password_generator()
    >>> passwords['redis'] = 'some-hardcoded-password'
    >>> passwords['redis']
    'some-hardcoded-password'
    >>> re.match(r'[0-9a-zA-Z]{16}', passwords['db']) is None
    False
    >>> passwords['db'] == passwords['db']
    True
    >>> passwords['cache'] == passwords['db']
    False
    """
    return collections.defaultdict(
        lambda: ''.join(random.SystemRandom().choice(characters) for _ in range(length)))


# TODO implement githash class with interface: hg.python.org/cpython/file/3.4/Lib/hashlib.py
# TODO add length optional argument to implement chunk'ed update()
def githash(path_or_data, encoding='utf-8', is_path=False, chunk_size=None):
    """
    Return the blob of some data.

    This is how Git `calculates <http://stackoverflow.com/questions/552659>`_
    the SHA1 for a file (or, in Git terms, a "blob")::

        sha1('blob ' + filesize + (the byte 0) + data)

    **Example usage**

    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> githash('')
    'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391'
    >>> githash('give me some hash please')
    'abdd1818289725c072eff0f5ce185457679650be'
    >>> githash('et ça fonctionne !\\n')
    '91de5baf6aaa1af4f662aac4383b27937b0e663d'
    >>> githash(directory / '..' / 'setup.py', is_path=True)
    '388b2b5ff8dc37a4f3f8997ff1ca47368a97d17b'
    >>> githash(directory / '..' / 'setup.py', is_path=True, chunk_size=256)
    '388b2b5ff8dc37a4f3f8997ff1ca47368a97d17b'
    """
    hasher = hashlib.sha1()
    if is_path:
        hasher.update((f'blob {os.path.getsize(path_or_data)}\0').encode('utf-8'))
        for data_bytes in filesystem.get_bytes(path_or_data, encoding, is_path, chunk_size):
            hasher.update(data_bytes)
    else:
        data_bytes = next(filesystem.get_bytes(path_or_data, encoding, is_path, chunk_size=None))
        hasher.update((f'blob {len(data_bytes)}\0').encode('utf-8'))
        hasher.update(data_bytes)
    return hasher.hexdigest()


def guess_algorithm(checksum, algorithms=None, unique=False):  # pylint:disable=redefined-outer-name
    """
    Guess the algorithms that have produced the checksum, based on its size.

    **Example usage**

    >>> algorithms = ('md4', 'md5', 'sha256', 'sha512', 'whirlpool')
    >>> long_checksum = '43d92a466b57e3744532eab7d760708028a7562d9678f6762bf341f29b921e42'
    >>> short_checksum = '2b31de8940dfd3286f70c316f701a54a'
    >>> guess_algorithm('', algorithms)
    set()
    >>> {a.name for a in guess_algorithm(long_checksum, algorithms)}
    {'sha256'}
    >>> guess_algorithm(long_checksum, algorithms, unique=True).name
    'sha256'
    >>> guess_algorithm(short_checksum, algorithms, unique=True) is None
    True

    Following examples depends of your system, so they are disabled::

    >> {a.name for a in guess_algorithm(short_checksum)}
    {'md4', 'md5'}
    >> {a.name for a in guess_algorithm(long_checksum))}
    {'sha256'}
    """
    digest_size = len(checksum) / 2
    if algorithms:
        algorithms = [hashlib.new(a) if isinstance(a, str) else a for a in algorithms]
    else:
        try:
            algorithms = [hashlib.new(a) for a in hashlib.algorithms_available if a.lower() == a]
        except AttributeError as e:
            raise NotImplementedError(
                "Your version of hashlib doesn't implement algorithms_available") from e
    digest_size_to_algorithms = collections.defaultdict(set)
    for algorithm in algorithms:
        digest_size_to_algorithms[algorithm.digest_size].add(algorithm)
    possible_algorithms = digest_size_to_algorithms[digest_size]
    if unique:
        return possible_algorithms.pop() if len(possible_algorithms) == 1 else None
    return possible_algorithms



import ezodf, pandas as pd


def map_dict(df, key_column, value_column, dropna=True):
    df = df[[key_column, value_column]]
    if dropna:
        df = df.dropna()
    return df.set_index(key_column)[value_column].to_dict()


def read_ods(filename, sheet_no, header_pos=0):
    tab = ezodf.opendoc(filename=filename).sheets[sheet_no]
    return pd.DataFrame({
        col[header_pos].value: [cell.value for cell in col[header_pos + 1:]]
        for col in tab.columns()
    })



from django import dispatch as _dispatch

from pytoolbox.django.models import utils as _utils

__all__ = ['InstanceSignal', 'post_state_transition']


class InstanceSignal(_dispatch.Signal):

    def __init__(self, providing_args=None, use_caching=False):
        providing_args = providing_args or ['instance']
        if 'instance' not in providing_args:
            providing_args.insert(0, 'instance')
        super().__init__(providing_args, use_caching)

    def send(self, sender=None, **named):
        return super().send(_utils.get_base_model(sender or named['instance']), **named)

    def send_robust(self, sender=None, **named):
        return super().send(_utils.get_base_model(sender or named['instance']), **named)


post_state_transition = InstanceSignal(providing_args=['previous_state', 'args', 'kwargs'])



"""
Mix-ins for building your own test cases.
"""

import os

from django.contrib.sites.models import Site
from django.core.management import call_command
from django.db import connections, DEFAULT_DB_ALIAS
from django.test import TransactionTestCase
from django.test.utils import CaptureQueriesContext

try:
    from django.urls import resolve, reverse
except ImportError:
    # For Django < 2.0
    from django.core.urlresolvers import resolve, reverse

from pytoolbox import module

_all = module.All(globals())


class _AssertNumQueriesInContext(CaptureQueriesContext):

    def __init__(self, test_case, num_range, connection):
        self.test_case = test_case
        self.range = num_range
        super().__init__(connection)

    def __exit__(self, exc_type, exc_value, traceback):
        super().__exit__(exc_type, exc_value, traceback)
        if exc_type is None:
            executed = len(self)
            self.test_case.assertIn(
                executed,
                self.range,
                f'{executed} queries executed, {self.range} expected{os.linesep}'
                f'Captured queries were:{os.linesep}'
                f"{os.linesep.join(query['sql'] for query in self.captured_queries)}")


class ClearSiteCacheMixin(object):

    def clear_site_cache(self):
        Site.objects.clear_cache()

    def setUp(self):
        self.clear_site_cache()
        super().setUp()

    def assertNumQueries(self, *args, **kwargs):
        self.clear_site_cache()
        return super().assertNumQueries(*args, **kwargs)


class FixFlushMixin(object):

    def _fixture_teardown(self):
        """
        Fix TransactionTestCase tear-down by enabling TRUNCATE CASCADE. Issue with Django 1.8a1.
        """
        assert isinstance(self, TransactionTestCase)
        for db_name in self._databases_names(include_mirrors=False):
            call_command(
                'flush',
                verbosity=0,
                interactive=False,
                database=db_name,
                reset_sequences=False,
                allow_cascade=True,
                inhibit_post_migrate=self.available_apps is not None)


class FormWizardMixin(object):

    def assertWizardSteps(self, response, **kwargs):
        for key, value in kwargs.items():
            self.assertEqual(getattr(response.context['wizard']['steps'], key), value, msg=key)

    def post_wizard(self, url, step, data=None, raw_data=None, **kwargs):
        from formtools.wizard.views import normalize_name
        name = normalize_name(resolve(reverse(url)).func.__name__)
        step_data = {f'{step}-{k}': v for k, v in data.items()} if data else {}
        step_data[f'{name}-current_step'] = step
        step_data.update(raw_data or {})
        return self.post(url, step_data, **kwargs)


class QueriesMixin(object):

    def assertNumQueriesIn(self, num_range, func=None, *args, **kwargs):
        connection = connections[kwargs.pop('using', DEFAULT_DB_ALIAS)]
        context = _AssertNumQueriesInContext(self, num_range, connection)
        if func is None:
            return context
        with context:
            func(*args, **kwargs)


class UrlMixin(object):

    def resolve(self, value, qs=None, urlconf=None, args=None, kwargs=None, current_app=None):
        if isinstance(value, str) and '/' in value:
            url = value
        elif hasattr(value, 'get_absolute_url'):
            url = value.get_absolute_url()
        else:
            url = reverse(value, urlconf, args, kwargs, current_app)
        return url if qs is None else '{0}?{1}'.format(url, qs)


class RestAPIMixin(UrlMixin):

    def _call(
        self,
        method,
        url,
        data,
        status,
        qs=None,
        urlconf=None,
        args=None,
        kwargs=None,
        current_app=None,
        msg=lambda r: getattr(r, 'data', r),
        **call_kwargs
    ):
        url = self.resolve(url, qs, urlconf, args, kwargs, current_app)
        response = getattr(self.client, method)(url, data, **call_kwargs)
        self.assertEqual(response.status_code, status, msg(response))
        return response

    def delete(self, url, data=None, status=204, **kwargs):
        return self._call('delete', url, data, status, **kwargs)

    def get(self, url, data=None, status=200, **kwargs):
        return self._call('get', url, data, status, **kwargs)

    def patch(self, url, data, status=200, **kwargs):
        return self._call('patch', url, data, status, **kwargs)

    def post(self, url, data, status=201, **kwargs):
        return self._call('post', url, data, status, **kwargs)

    def put(self, url, data, status=200, **kwargs):
        return self._call('put', url, data, status, **kwargs)


__all__ = _all.diff(globals())



"""
Meta-classes for enhancing your models.
"""

import abc

from django.db import models

_all__ = ['ABCModelMeta']


class ABCModelMeta(abc.ABCMeta, type(models.Model)):
    """
    Meta-class for building an abstract Model with abstract methods, properties, ...

    **Example usage**

    >> class AbstractModel(models.Model):
    .. __metaclass__ = AbstractModelMeta
    ..
    .. class Meta:
    ..     abstract = True
    """



import collections, inspect, signal

from . import exceptions

__all__ = ['handlers_by_signal', 'propagate_handler', 'register_handler', 'register_callback']

handlers_by_signal = collections.defaultdict(list)


def propagate_handler(signum, frame):
    errors = {}
    for handler in reversed(handlers_by_signal[signum]):
        try:
            handler(signum, frame)
        except Exception as e:  # pylint:disable=broad-except
            errors[handler] = e
    if errors:
        raise RuntimeError(errors)


def register_handler(signum, handler, append=True, reset=False):
    old_handler = signal.getsignal(signum)
    signal.signal(signum, propagate_handler)
    if inspect.isfunction(old_handler) and old_handler is not propagate_handler:
        handlers_by_signal[signum].append(old_handler)
    handlers = handlers_by_signal[signum]
    if not append and handlers:
        raise exceptions.MultipleSignalHandlersError(signum=signum, handlers=handlers)
    if reset:
        try:
            handlers.clear()
        except AttributeError:
            # < Python 3.3
            del handlers[:]
    handlers.append(handler)


def register_callback(signum, callback, append=True, reset=False, args=None, kwargs=None):
    return register_handler(
        signum,
        lambda s, f: callback(*(args or []), **(kwargs or {})),
        append,
        reset)



from urllib.parse import urlsplit, urlunsplit

from pytoolbox import string

__all__ = ['with_subdomain']


def with_subdomain(url, subdomain=None):
    """
    Return the ``url`` with the sub-domain replaced with ``subdomain``.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>> eq = asserts.equal
    >>> asserts.equal(
    ...     with_subdomain('http://app.website.com/page'),
    ...     'http://website.com/page')
    >>> asserts.equal(
    ...     with_subdomain('http://app.website.com/page', 'help'),
    ...     'http://help.website.com/page')
    >>> asserts.equal(
    ...     with_subdomain('https://app.website.com#d?page=1', 'help'),
    ...     'https://help.website.com#d?page=1')
    """
    import tldextract
    components = list(urlsplit(url))
    components[1] = string.filterjoin((subdomain, ) + tldextract.extract(components[1])[1:], '.')
    return urlunsplit(components)



"""
Mix-ins for building your own `Django Form Tools <https://github.com/django/django-formtools>`_
powered views.
"""

from pytoolbox.django import forms

__all__ = ['CrispyFormsMixin', 'DataTableViewCompositionMixin', 'SerializeStepInstanceMixin']


class CrispyFormsMixin(object):

    def get_context_data(self, form, **kwargs):
        """Add the management form to the form for working with crispy forms."""
        from crispy_forms import layout
        context = super().get_context_data(form=form, **kwargs)
        context['wizard']['form'].helper.layout.append(
            layout.HTML(context['wizard']['management_form']))
        context['form'] = context['wizard']['form']
        return context


class DataTableViewCompositionMixin(object):
    """Compose the wizard with some tables views."""

    table_view_classes = {}

    def get(self, request, *args, **kwargs):
        """Retrieve the table view and delegate AJAX to the table view."""
        if self.request.is_ajax():
            view = self.get_table_view()
            if view:
                return view.get_ajax(request, *args, **kwargs)
        return super().get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        """Update the context with the context returned by the table view."""
        context = super().get_context_data(**kwargs)
        view = self.get_table_view()
        if view:
            context.update(view.get_context_data())
        return context

    def get_table_view(self):
        """Return an instance of the datatable-view for current step, defaults to None."""
        view_class = self.table_view_classes.get(self.steps.current)
        if view_class:
            view = view_class()
            view.object_list = None
            view.request = self.request
            return view


class SerializeStepInstanceMixin(object):

    serialized_instance_form_class = forms.SerializedInstanceForm
    serialized_instances_key = 'serialized-instances'

    @property
    def serialized_instances(self):
        try:
            return self.storage.extra_data[self.serialized_instances_key]
        except KeyError:
            value = self.storage.extra_data[self.serialized_instances_key] = {}
            return value

    def serialize_step_instance(self, form, step=None):
        self.serialized_instances[step or self.steps.current] = \
            self.serialized_instance_form_class.serialize(form.save())

    # WizardView "Standard Methods"

    def get_form(self, step=None, *args, **kwargs):
        if step is None:
            step = self.steps.current
        if step in self.serialized_instances.keys():
            self.form_list[step] = self.serialized_instance_form_class
        return super().get_form(step, *args, **kwargs)

    def get_form_kwargs(self, step):
        form_kwargs = super().get_form_kwargs(step)
        serialized_instance = self.serialized_instances.get(step, None)
        if serialized_instance:
            form_kwargs.update(serialized_instance)
        return form_kwargs



"""
Mix-ins for building your own query-sets.
"""

import functools

from django.db import transaction

from pytoolbox import module

_all = module.All(globals())


class AtomicGetUpdateOrCreateMixin(object):

    savepoint = False

    def get_or_create(self, defaults=None, **kwargs):
        with transaction.atomic(savepoint=self.savepoint):
            return super().get_or_create(defaults=defaults, **kwargs)

    def update_or_create(self, defaults=None, **kwargs):
        with transaction.atomic(savepoint=self.savepoint):
            return super().update_or_create(defaults=defaults, **kwargs)


class AtomicGetRestoreOrCreateMixin(object):

    savepoint = False

    def get_restore_or_create(self, *args, **kwargs):
        with transaction.atomic(savepoint=self.savepoint):
            return super().get_restore_or_create(*args, **kwargs)


class CreateModelMethodMixin(object):

    def create(self, *args, **kwargs):
        if hasattr(self.model, 'create'):
            return self.model.create(*args, **kwargs)
        return super().create(*args, **kwargs)
    create.alters_data = True


class StateMixin(object):
    """
    Generate on the fly utility query-set filtering methods to a model using a
    :class:`pytoolbox.states.StateEnum` to implement its own state machine. Then you can use
    something like ``Model.objects.ready_or_canceled(inverse=True)`` to exclude models in state
    READY or CANCELED.

    This mixin requires the following to work:

    * Add a `states` attribute to your model class set to the states class you defined earlier.
    * Add a `state` field to the model for saving instance state in database.
    """

    _skip_names = frozenset(['__getstate__', 'model'])

    def __getattr__(self, name):
        # avoid strange infinite recursion with defer()
        if name not in self._skip_names:
            all_states = set()
            for name in name.split('_or_'):
                states = self.model.states.get(name)
                if not states:
                    raise AttributeError
                method = all_states.add if isinstance(states, str) else all_states.update
                method(states)
            return functools.partial(self.in_states, all_states)
        raise AttributeError

    def in_states(self, states, inverse=False):
        """Filter query set to include instances in `states`."""
        method = self.exclude if inverse else self.filter
        return method(state__in={states} if isinstance(states, str) else states)


__all__ = _all.diff(globals())



"""
Extra storages and mix-ins for building your own storages.
"""

import logging, time

from django.core.files.storage import FileSystemStorage

from pytoolbox import module

_all = module.All(globals())

logger = logging.getLogger(__name__)


class ExpressTemporaryFileMixin(object):

    def _save(self, name, content):
        """
        Save the temporary file to the storage.
        Set temporary file path to allow using rename instead of chunked copy when possible.

        See what happens in super()._save():
            https://github.com/django/django/blob/master/django/core/files/storage.py
        """
        start_time = time.time()
        if hasattr(content.file, 'temporary_file_path'):
            content.temporary_file_path = lambda: content.file.temporary_file_path()
        result = super()._save(name, content)
        logger.debug(f'Saved protected file "{name}" in {time.time() - start_time:.2g} seconds')
        return result


class OverwriteMixin(object):
    """
    Update get_available_name to remove any previously stored file (if any) before returning the
    name.
    """

    def get_available_name(self, name):
        self.delete(name)
        return name


class OverwriteFileSystemStorage(OverwriteMixin, FileSystemStorage):
    """A file-system based storage that let overwrite files with the same name."""


__all__ = _all.diff(globals())



from pytoolbox import decorators
from . import brand, equipment  # pylint:disable=unused-import

__all__ = ['Camera']


class Camera(equipment.Equipement):

    brand_class = brand.Brand

    @property
    def brand(self):
        return self.brand_class(self.metadata['Exif.Image.Make'].data)

    @property
    def _model(self):
        return self.metadata['Exif.Image.Model'].data

    @decorators.cached_property
    def tags(self):
        return {k: t for k, t in self.metadata.tags.items() if 'camera' in t.label.lower()}



#!/usr/bin/env python

# **************************************************************************************************
#                              PYTOOLBOX - TOOLBOX FOR PYTHON SCRIPTS
#
#  Main Developer : David Fischer (david.fischer.ch@gmail.com)
#  Copyright      : Copyright (c) 2012-2020 David Fischer. All rights reserved.
#
# **************************************************************************************************
#
# This file is part of David Fischer's pytoolbox Project.
#
# This project is free software: you can redistribute it and/or modify it under the terms of the
# EUPL v. 1.1 as provided by the European Commission. This project is distributed in the hope that
# it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#
# See the European Union Public License for more details.
#
# You should have received a copy of the EUPL General Public License along with this project.
# If not, see he EUPL licence v1.1 is available in 22 languages:
#     22-07-2013, <https://joinup.ec.europa.eu/software/page/eupl/licence-eupl>
#
# Retrieved from https://github.com/davidfischer-ch/pytoolbox.git

import itertools, setuptools, subprocess, sys
from pathlib import Path
from setuptools.command import develop, install

import pytoolbox

if sys.argv[-1] == 'test':
    sys.exit('Run pytest instead.')

install_requires = [
    'argparse',
    'pyaml',
    'python-magic',
    'pytz',
    'requests',
    'termcolor'
]

try:
    import grp
    _ = grp
except ImportError:
    # Required on Windows
    install_requires.append('python-magic-bin')

extras_require = {
    'atlassian': [
        'jira'
    ],
    'aws': [
        'boto3'
    ],
    'django': [
        'django'
    ],
    'django_filter': [
        'django-filter'
    ],
    'django_formtools': [
        'django-formtools'
    ],
    'flask': [
        'flask'
    ],
    'imaging': [
        'pillow'
    ],
    'jinja2': [
        'jinja2'
    ],
    'network': [
        'tldextract'
    ],
    'pandas': [
        'ezodf',
        'lxml',
        'pandas'
    ],
    'rest_framework': [
        'django-oauth-toolkit',
        'djangorestframework>=3'
    ],
    'selenium': [
        'selenium'
    ],
    'smpte2022': [
        'fastxor'
    ],
    'vision': [
        'dlib',
        'keras',
        'numpy',
        'opencv-python',
        'tensorflow'
    ],
    'voluptuous': [
        'voluptuous'
    ]
}

features = ['all'] + sorted(extras_require) + ['doc', 'test']

labels = {
    # Features
    'all': 'All Modules',
    'atlassian': 'Atlassian',
    'aws': 'AWS',
    'django': 'Django',
    'django_filter': 'Django Filter',
    'django_formtools': 'Django Form Tools',
    'flask': 'Flask',
    'imaging': 'Imaging',
    'jinja2': 'Jinja2',
    'logging': 'Logging',
    'network': 'Networking',
    'pandas': 'Pandas',
    'rest_framework': 'Django REST Framework',
    'selenium': 'Selenium',
    'smpte2022': 'SMPTE-2022',
    'unittest': 'Unit Test',
    'vision': 'Vision',
    'voluptuous': 'Voluptuous',

    # Actions
    'doc': 'Pytoolbox Docs',
    'test': 'Pytoolbox Tests',
}

extras_require.update({
    'all': sorted(set(itertools.chain.from_iterable(extras_require.values()))),
    'doc': [
        'sphinx>=1.3.1',
        'sphinx_rtd_theme'
    ],
    'test': [
        'coverage>=6.4.4<7',       # 01-09-2022 Released 16-08-2022
        'flake8>=4,<5',            # 01-09-2022 Released 11-10-2021
                                   # Bug https://github.com/tholo/pytest-flake8/issues/87
                                   # Fix https://github.com/tholo/pytest-flake8/pull/88/files
        'pylint>=2.15.0<3',        # 01-09-2022 Released 26-08-2022
        'pytest>=7.1.2<8',         # 01-09-2022 Released 23-04-2022
        'pytest-cov>=3.0.0<4',     # 01-09-2022 Released 04-10-2021
        'pytest-flake8>=1.1.1<2',  # 01-09-2022 Released 18-03-2022
        'pytest-pylint>=0.18.0<1'  # 01-09-2022 Released 09-11-2020
    ]
})


def get_command_with_extras(cls):

    class WithExtra(cls):

        user_options = [
            *cls.user_options,
            *[
                (
                    f"extra-{feature.replace('_', '-')}", None,
                    f'Install dependencies for {labels[feature]}.'
                ) for feature in features
            ]
        ]

        boolean_options = [
            *getattr(cls, 'boolean_options', []),
            'extra-all',
            *[o[0] for o in user_options]
        ]

        def initialize_options(self):
            cls.initialize_options(self)
            for feature in features:
                setattr(self, f'extra_{feature}', None)

        def finalize_options(self):
            cls.finalize_options(self)
            for feature in features:
                if getattr(self, f'extra_{feature}'):
                    print(f'Enable dependencies for {labels[feature]}')
                    self.distribution.install_requires.extend(
                        self.distribution.extras_require[feature])

    return WithExtra


class DocsCommand(setuptools.Command):  # pylint:disable=duplicate-code

    description = 'Generate documentation using Sphinx'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        from pytoolbox import filesystem
        project = Path(__file__).resolve().parent
        source = project / 'docs' / 'source'

        # Cleanup previously generated restructured files
        for path in filesystem.find_recursive(source, r'^pytoolbox.*\.rst$', regex=True):
            filesystem.remove(path)

        subprocess.run([
            'sphinx-apidoc',
            '--force',
            '--module-first',
            '--separate',
            '-o', source,
            project / 'pytoolbox'
        ], check=True)
        filesystem.remove(project / 'docs' / 'build' / 'html', recursive=True)
        subprocess.run(['make', 'html'], cwd=project / 'docs', check=True)


with open('README.rst', encoding='utf-8') as f:
    long_description = f.read()

setuptools.setup(
    cmdclass={
        'docs': DocsCommand,
        'develop': get_command_with_extras(develop.develop),
        'install': get_command_with_extras(install.install)
    },
    name='pytoolbox',
    python_requires='>=3.9',
    version=pytoolbox.__version__,
    packages=setuptools.find_packages(exclude=['tests']),
    package_data={
        'pytoolbox.ai.vision.face.detect': ['data/*'],
        'pytoolbox.ai.vision.face.recognize': ['data/*']
    },
    extras_require=extras_require,
    install_requires=install_requires,
    tests_require=extras_require['test'],

    # Meta-data for upload to PyPI
    author='David Fischer',
    author_email='david@fisch3r.net',
    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Framework :: Django',
        'Framework :: Flask',
        'License :: OSI Approved :: European Union Public Licence 1.1 (EUPL 1.1)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    description='Toolbox for Python scripts',
    keywords=[
        'celery',
        'ffmpeg',
        'django',
        'flask',
        'json',
        'juju',
        'mock',
        'pillow',
        'rsync',
        'rtp',
        'selenium',
        'smpte 2022-1',
        'screen',
        'subprocess'
    ],
    license='EUPL 1.1',
    long_description=long_description,
    url='https://github.com/davidfischer-ch/pytoolbox'
)



__all__ = ['All']


class All(object):

    def __init__(self, globals_):
        self.init_keys = set(globals_.keys())

    def diff(self, globals_, to_type=list):
        new_keys = set(globals_.keys()) - self.init_keys
        return to_type(k for k in new_keys if k[0] != '_')



"""
Extra `fields <http://www.django-rest-framework.org/api-guide/fields/>`_ for building your own
`Django REST Framework <https://github.com/tomchristie/django-rest-framework>`_ powered API
`serializers <http://www.django-rest-framework.org/tutorial/1-serialization/>`_.
"""

from rest_framework import serializers

from pytoolbox.django.core.validators import EmptyValidator

__all__ = ['StripCharField']


class StripCharField(serializers.CharField):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.validators.append(EmptyValidator(message=self.error_messages['blank']))

    def to_internal_value(self, data):
        data = super().to_internal_value(data)
        return data.strip() if data else data



# Copyright 2015-2016 Carnegie Mellon University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Original: https://github.com/krasserm/face-recognition/blob/master/align.py

import bz2, tempfile

import cv2, dlib, numpy as np

from pytoolbox.ai.vision import utils

__all__ = ['DlibFaceDetector']

TEMPLATE = np.float32([
    (0.0792396913815, 0.339223741112), (0.0829219487236, 0.456955367943),
    (0.0967927109165, 0.575648016728), (0.122141515615, 0.691921601066),
    (0.168687863544, 0.800341263616), (0.239789390707, 0.895732504778),
    (0.325662452515, 0.977068762493), (0.422318282013, 1.04329000149),
    (0.531777802068, 1.06080371126), (0.641296298053, 1.03981924107),
    (0.738105872266, 0.972268833998), (0.824444363295, 0.889624082279),
    (0.894792677532, 0.792494155836), (0.939395486253, 0.681546643421),
    (0.96111933829, 0.562238253072), (0.970579841181, 0.441758925744),
    (0.971193274221, 0.322118743967), (0.163846223133, 0.249151738053),
    (0.21780354657, 0.204255863861), (0.291299351124, 0.192367318323),
    (0.367460241458, 0.203582210627), (0.4392945113, 0.233135599851),
    (0.586445962425, 0.228141644834), (0.660152671635, 0.195923841854),
    (0.737466449096, 0.182360984545), (0.813236546239, 0.192828009114),
    (0.8707571886, 0.235293377042), (0.51534533827, 0.31863546193),
    (0.516221448289, 0.396200446263), (0.517118861835, 0.473797687758),
    (0.51816430343, 0.553157797772), (0.433701156035, 0.604054457668),
    (0.475501237769, 0.62076344024), (0.520712933176, 0.634268222208),
    (0.565874114041, 0.618796581487), (0.607054002672, 0.60157671656),
    (0.252418718401, 0.331052263829), (0.298663015648, 0.302646354002),
    (0.355749724218, 0.303020650651), (0.403718978315, 0.33867711083),
    (0.352507175597, 0.349987615384), (0.296791759886, 0.350478978225),
    (0.631326076346, 0.334136672344), (0.679073381078, 0.29645404267),
    (0.73597236153, 0.294721285802), (0.782865376271, 0.321305281656),
    (0.740312274764, 0.341849376713), (0.68499850091, 0.343734332172),
    (0.353167761422, 0.746189164237), (0.414587777921, 0.719053835073),
    (0.477677654595, 0.706835892494), (0.522732900812, 0.717092275768),
    (0.569832064287, 0.705414478982), (0.635195811927, 0.71565572516),
    (0.69951672331, 0.739419187253), (0.639447159575, 0.805236879972),
    (0.576410514055, 0.835436670169), (0.525398405766, 0.841706377792),
    (0.47641545769, 0.837505914975), (0.41379548902, 0.810045601727),
    (0.380084785646, 0.749979603086), (0.477955996282, 0.74513234612),
    (0.523389793327, 0.748924302636), (0.571057789237, 0.74332894691),
    (0.672409137852, 0.744177032192), (0.572539621444, 0.776609286626),
    (0.5240106503, 0.783370783245), (0.477561227414, 0.778476346951)
])

TPL_MIN, TPL_MAX = np.min(TEMPLATE, axis=0), np.max(TEMPLATE, axis=0)
MINMAX_TEMPLATE = (TEMPLATE - TPL_MIN) / (TPL_MAX - TPL_MIN)


class DlibFaceDetector(object):
    """
    Use `dlib's landmark estimation
    <http://blog.dlib.net/2014/08/real-time-face-pose-estimation.html>`_
    to align faces.

    The alignment pre-process faces for input into a neural network.
    Faces are resized to the same size (such as 96x96) and transformed to make landmarks
    (such as the eyes and nose) appear at the same location on every image.

    Normalized landmarks:

    .. image:: ../images/dlib-landmark-mean.png
    """

    # Downloaded from http://dlib.net/files/
    DEFAULT_PREDICTOR = (
        'https://s3-eu-west-1.amazonaws.com/pytoolbox/ai/vision/face/detect/'
        'shape_predictor_68_face_landmarks.dat.bz2')

    PREDICTORS_DOWNLOAD_DIRECTORY = tempfile.gettempdir()

    #: Landmark indices.
    INNER_EYES_AND_BOTTOM_LIP = [39, 42, 57]
    OUTER_EYES_AND_NOSE = [36, 45, 33]

    def __init__(self, predictor=None):
        """
        Instantiate a :class:`DlibFaceDetector` object.

        :param predictor: The path to dlib's landmarks file.
        :type predictor: str
        """
        if predictor is None:
            predictor = self.DEFAULT_PREDICTOR
        self.predictor = self._load_predictor(predictor)
        self.detector = dlib.get_frontal_face_detector()  # pylint:disable=no-member

    def align(self, image, box, dimension=96, landmark_indices=None, landmarks=None):
        """
        Transform and align a face in an image.

        :param image: RGB image to process. Shape: (height, width, 3)
        :type image: numpy.ndarray
        :param box: Bounding box around the face to align.
        :type box: dlib.rectangle
        :type dimension: int
        :param landmark_indices: The indices to transform to.
        :type landmark_indices: list of ints
        :param dimension: The edge length in pixels of the square the image is resized to.
        :param landmarks: Detected landmark locations. \
                          Landmarks found on `box` if not provided.
        :type landmarks: list of (x,y) tuples
        :return: The aligned RGB image. Shape: (dimension, dimension, 3)
        :rtype: numpy.ndarray
        """
        if landmark_indices is None:
            landmark_indices = self.OUTER_EYES_AND_NOSE
        if landmarks is None:
            landmarks = self.find_landmarks(image, box)

        landmarks = np.float32(landmarks)
        landmark_indices = np.array(landmark_indices)

        H = cv2.getAffineTransform(       # pylint:disable=invalid-name,no-member
            landmarks[landmark_indices],  # pylint:disable=unsubscriptable-object
            dimension * MINMAX_TEMPLATE[landmark_indices])

        return cv2.warpAffine(image, H, (dimension, dimension))  # pylint:disable=no-member

    def extract_all_faces(self, image, dimension=96, landmark_indices=None):
        for box in self.get_all_faces_bounding_boxes(image):
            yield box, self.align(image, box, dimension, landmark_indices)

    def extract_largest_face(self, image, dimension=96, landmark_indices=None, skip_multi=False):
        box = self.get_largest_face_bounding_box(image, skip_multi=skip_multi)
        return box, (self.align(image, box, dimension, landmark_indices) if box else None)

    def find_landmarks(self, image, box):
        """
        Find the landmarks of a face.

        :param image: RGB image to process. Shape: (height, width, 3)
        :type image: numpy.ndarray
        :param box: Bounding box around the face to find landmarks for.
        :type box: dlib.rectangle
        :return: Detected landmark locations.
        :rtype: list of (x,y) tuples
        """
        points = self.predictor(image, box)
        return [(p.x, p.y) for p in points.parts()]

    def get_all_faces_bounding_boxes(self, image):
        """
        Find all face bounding boxes in an image.

        :param image: RGB image to process. Shape: (height, width, 3)
        :type image: numpy.ndarray
        :return: All face bounding boxes in an image.
        :rtype: dlib.rectangles
        """
        try:
            return self.detector(image, 1)
        except Exception as e:  # pylint:disable=broad-except
            print(f"Warning: {e}")
            # In rare cases, exceptions are thrown.
            return []

    def get_largest_face_bounding_box(self, image, skip_multi=False):
        """
        Find the largest face bounding box in an image.

        :param image: RGB image to process. Shape: (height, width, 3)
        :type image: numpy.ndarray
        :param skip_multi: Skip image if more than one face detected.
        :type skip_multi: bool
        :return: The largest face bounding box in an image, or None.
        :rtype: dlib.rectangle
        """
        faces = self.get_all_faces_bounding_boxes(image)
        if not skip_multi and len(faces) > 0 or len(faces) == 1:
            return max(faces, key=lambda rect: rect.width() * rect.height())
        return None

    @staticmethod
    def _load_predictor(predictor):
        predictor = utils.load_to_file(predictor)
        if predictor.endswith('.bz2'):
            decompressor = bz2.BZ2Decompressor()
            with open(predictor, 'rb') as source, tempfile.NamedTemporaryFile('wb') as target:
                target.write(decompressor.decompress(source.read()))
                return dlib.shape_predictor(target.name)  # pylint:disable=no-member
        else:
            return dlib.shape_predictor(predictor)  # pylint:disable=no-member



# pylint:disable=invalid-name

import functools

from pytoolbox import module

_all = module.All(globals())

from PIL import Image  # noqa pylint:disable=wrong-import-position

TRANSPOSE_SEQUENCES = {
    None: [],
    1: [],
    2: [Image.FLIP_LEFT_RIGHT],
    3: [Image.ROTATE_180],
    4: [Image.FLIP_TOP_BOTTOM],
    5: [Image.FLIP_LEFT_RIGHT, Image.ROTATE_90],
    6: [Image.ROTATE_270],
    7: [Image.FLIP_TOP_BOTTOM, Image.ROTATE_90],
    8: [Image.ROTATE_90]
}


def get_orientation(image, orientation_tag=0x0112, no_exif_default=None, no_key_default=None):
    exif = getattr(image, '_getexif', lambda: None)()
    try:
        return no_exif_default if exif is None else exif[orientation_tag]
    except KeyError:
        return no_key_default


def apply_orientation(  # pylint:disable=dangerous-default-value
    image,
    get_orientation=get_orientation,  # pylint:disable=redefined-outer-name
    sequences=TRANSPOSE_SEQUENCES
):
    """Credits: https://stackoverflow.com/questions/4228530/pil-thumbnail-is-rotating-my-image."""
    orientation = get_orientation(image)
    return functools.reduce(lambda i, op: i.transpose(op), sequences.get(orientation, []), image)


def open(file_or_path):  # pylint:disable=redefined-builtin
    image = Image.open(file_or_path)
    try:
        image.load()
    except IOError as e:
        if 'truncated' not in str(e):
            raise
    return image


def remove_metadata(image, keys=('exif', ), inplace=False):
    image = image if inplace else image.copy()
    for key in keys:
        if key in image.info:
            image.info[key] = b''
    return image


def remove_transparency(image, background=(255, 255, 255)):
    """
    Return a RGB image with an alpha mask applied to picture + background.
    If image is already in RGB, then its a no-op.
    """
    if image.mode == 'RGB':
        return image  # No-op
    alpha = image.convert('RGBA').getchannel('A')
    new_image = Image.new('RGB', image.size, background)
    new_image.paste(image.convert('RGB'), mask=alpha)
    return new_image


def save(image, *args, **kwargs):
    kwargs.setdefault('exif', image.info.get('exif', b''))
    return image.save(*args, **kwargs)


__all__ = _all.diff(globals())



"""
Extra fields for your models.
"""

import math, os

from django.conf import settings
from django.db import models
from django.core import validators as dj_validators
from django.db.models.fields import files
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _

from pytoolbox import module
from pytoolbox.django.core import validators
from . import mixins

_all = module.All(globals())


# Char & Text

class CharField(mixins.OptionsMixin, mixins.NullifyMixin, models.CharField):
    """A CharFiled with OptionsMixin and NullifyMixin applied."""


class TextField(mixins.OptionsMixin, mixins.NullifyMixin, models.TextField):
    """A TextField with OptionsMixin and NullifyMixin applied."""


class StripCharField(mixins.OptionsMixin, mixins.StripMixin, mixins.NullifyMixin, models.CharField):
    """A CharFiled with OptionsMixin, StripMixin and NullifyMixin applied."""


class StripTextField(mixins.OptionsMixin, mixins.StripMixin, mixins.NullifyMixin, models.TextField):
    """A TextField with OptionsMixin, StripMixin and NullifyMixin applied."""


class ExtraChoicesField(StripCharField):

    def __init__(self, verbose_name=None, extra_choices=None, **kwargs):
        self.extra_choices = extra_choices or []
        super().__init__(verbose_name=verbose_name, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(ExtraChoicesField, self).deconstruct()
        if self.extra_choices:
            kwargs['extra_choices'] = self.extra_choices
        return name, path, args, kwargs

    def validate(self, value, model_instance):
        choices = self._choices
        try:
            self._choices = list(self.choices) + list(self.extra_choices)
            return super(ExtraChoicesField, self).validate(value, model_instance)
        finally:
            self._choices = choices


# Date and time

class CreatedAtField(mixins.OptionsMixin, models.DateTimeField):

    default_options = {'default': now, 'editable': False, 'verbose_name': _('Created at')}


class UpdatedAtField(mixins.OptionsMixin, models.DateTimeField):

    default_options = {'auto_now': True, 'editable': False, 'verbose_name': _('Updated at')}


# Miscellaneous

class CreatedByField(mixins.OptionsMixin, models.ForeignKey):

    default_options = {'to': settings.AUTH_USER_MODEL, 'editable': False}


class MD5ChecksumField(StripCharField):

    default_error_messages = {'invalid': _('Enter a valid MD5 checksum')}
    default_options = {'max_length': 32}
    default_validators = [validators.MD5ChecksumValidator()]


class MoneyField(mixins.OptionsMixin, models.DecimalField):

    def __init__(self, max_value, decimal_places=2, **kwargs):
        self.max_value = max_value
        super().__init__(
            decimal_places=decimal_places,
            max_digits=int(math.log10(max_value)) + 3,
            validators=[
                dj_validators.MinValueValidator(0),
                dj_validators.MaxValueValidator(max_value)
            ],
            **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(MoneyField, self).deconstruct()
        kwargs.pop('decimal_places', None)
        kwargs.pop('max_digits', None)
        kwargs.pop('validators', None)
        return name, path, [self.max_value], kwargs


class URLField(StripCharField, models.URLField):
    """An URLFIeld with StripCharField behavior inside."""

    default_options = {'max_length': 8000}  # http://tools.ietf.org/html/rfc7230#section-3.1.1


# Storage

class FieldFile(files.FieldFile):

    @property
    def basename(self):
        return os.path.basename(self.name) if self else None

    @basename.setter
    def basename(self, value):
        # TODO use storage.get_valid_name
        self.name = self.field.upload_to(self.instance, os.path.basename(value))
        setattr(self.instance, self.field.name, self.name)

    @property
    def exists(self):
        return bool(self) and self.storage.exists(self.name)


class FileField(mixins.OptionsMixin, models.FileField):
    """A FileField with OptionsMixin applied."""

    attr_class = FieldFile


__all__ = _all.diff(globals())



"""
Some utilities related to the model layer.
"""

from pytoolbox import module

_all = module.All(globals())


def get_base_model(cls_or_instance):
    return cls_or_instance._meta.proxy_for_model or cls_or_instance._meta.model


def get_related_manager(cls_or_instance, field):
    return get_related_model(cls_or_instance, field)._default_manager


def get_related_model(cls_or_instance, field):
    if field == 'pk':
        field = cls_or_instance._meta.pk.attname.rstrip('_id')
    return cls_or_instance._meta.get_field(field).related_model


def get_content_type_dict(instance):
    """Return a dictionary with the serialized content type and private key of given instance."""
    from django.contrib.contenttypes import models as ct_models
    content_type = ct_models.ContentType.objects.get_for_model(instance.__class__)
    return {'app_label': content_type.app_label, 'model': content_type.model, 'pk': instance.pk}


def get_instance(app_label, model, pk):
    """Return an instance given its app_label, model name and private key."""
    from django.contrib.contenttypes import models as ct_models
    model = ct_models.ContentType.objects.get(app_label=app_label, model=model)
    return model.get_object_for_this_type(pk=pk)


def try_get_field(instance, field_name):
    try:
        return getattr(instance, field_name)
    except Exception as e:
        if e.__class__.__name__ != 'RelatedObjectDoesNotExist':
            raise


__all__ = _all.diff(globals())



"""
Throttling classes implementing various throttling policies.
"""

import time

from . import module
from .datetime import total_seconds
from .types import Missing

_all = module.All(globals())


class TimeThrottle(object):
    """
    Time based throttling class.

    >>> import datetime
    >>> def slow_range(*args):
    ...     for i in range(*args):
    ...         time.sleep(0.5)
    ...         yield i
    >>> t1, t2 = (TimeThrottle(t) for t in (datetime.timedelta(minutes=1), 0.2))
    >>> list(t1.throttle_iterable((i, i) for i in range(10)))
    [(0, 0), (9, 9)]
    >>> list(t2.throttle_iterable(slow_range(3)))
    [0, 1, 2]
    """

    def __init__(self, min_time_delta):
        self.min_time_delta = total_seconds(min_time_delta)
        self.previous_time = None

    def is_throttled(self):
        """Return a boolean indicating if you should throttle."""
        if not self.previous_time:
            self._update()
            return False
        if time.time() - self.previous_time >= total_seconds(self.min_time_delta):
            self._update()
            return False
        return True

    def throttle_iterable(self, objects, callback=lambda o: None):
        """
        Consume and skips some objects to yield them at defined `min_delay`. First and last objects
        are always returned.

        * Set `callback` to a callable with the signature ``is_throttled_args = callback(object)``.
          Used by subclasses.
        """
        current_object = Missing
        for obj in objects:
            current_object = obj
            args = callback(obj) or ()
            if not self.is_throttled(*args):
                current_object = Missing
                yield obj
        if current_object is not Missing:
            yield current_object

    def _update(self):
        self.previous_time = time.time()


class TimeAndRatioThrottle(TimeThrottle):
    """
    Time and ratio based throttling class.

    >>> import datetime
    >>> def slow_range(*args):
    ...     for i in range(*args):
    ...         time.sleep(0.5)
    ...         yield i
    >>> t1, t2 = (TimeAndRatioThrottle(0.3, t, 10*t) for t in (datetime.timedelta(minutes=1), 0.4))
    >>> list(t1.throttle_iterable(list(range(9)), lambda i: [i/9]))
    [0, 8]
    >>> list(t2.throttle_iterable(slow_range(9), lambda i: [i/9]))
    [0, 3, 6, 8]
    """

    def __init__(self, min_ratio_delta, min_time_delta, max_time_delta):
        super().__init__(min_time_delta)
        self.min_ratio_delta = total_seconds(min_ratio_delta)
        self.max_time_delta = total_seconds(max_time_delta)
        self.previous_ratio = 0

    def is_throttled(self, ratio):  # pylint:disable=arguments-differ
        """Return a boolean indicating if you should throttle."""
        if not self.previous_time:
            self._update(ratio)
            return False
        ratio_delta = ratio - self.previous_ratio
        time_delta = time.time() - self.previous_time
        if (
            ratio_delta > self.min_ratio_delta
            and time_delta > self.min_time_delta
            or time_delta > self.max_time_delta
        ):
            self._update(ratio)
            return False
        return True

    def _update(self, ratio):  # pylint:disable=arguments-differ
        super()._update()
        self.previous_ratio = ratio


__all__ = _all.diff(globals())



"""
    :file:`apps.py` ::

        from django import apps
        from django.utils.translation import gettext_lazy as _

        from . import signals

        __all__ = ('MyApp', )


        class MyAppConfig(apps.AppConfig):
            name = 'myapp'
            verbose_name = _('My Application')

            def ready(self):
                signals.connect(self)

    :file:`signals.py` ::

        from django.db.models import signals as dj_signals
        from django.db.backends import signals as dj_db_signals
        from pytoolbox.django.signals import (
            create_site, setup_postgresql_hstore_extension, strip_strings_and_validate_model
        )

        # ...

        def connect(config):
            '''Connect signal handlers to signals.'''
            dj_db_signals.connection_created.connect(setup_postgresql_hstore_extension)
            dj_signals.post_migrate.connect(create_site, sender=config)
            dj_signals.pre_save.connect(
                strip_strings_and_validate_model, sender=settings.AUTH_USER_MODEL)
"""

import logging

from django.conf import settings
from django.db.models import fields
from django.db.models.fields.files import FileField

__all__ = [
    'logger',
    'clean_files_delete_handler',
    'create_site',
    'setup_postgresql_hstore_extension',
    'strip_strings_and_validate_model'
]

logger = logging.getLogger(__name__)


def clean_files_delete_handler(instance, signal, **kwargs):
    """
    Remove the files of the instance's file fields when it is removed from the database.

    Simply use ``post_delete.connect(clean_files_delete_handler, sender=<your_model_class>)``

    .. warning::
        This function remove the file without worrying about any other instance using this file !

    .. note::
        Project `django-cleanup <https://github.com/un1t/django-cleanup>`_
        is a more complete alternative.
    """
    for field in kwargs['sender']._meta.fields:
        if isinstance(field, FileField):
            file_field = getattr(instance, field.name)
            if file_field.path:
                file_field.delete(save=False)


def create_site(sender, **kwargs):
    """
    Ensure the site name and domain is well configured.

    Some alternative:

    * Loading an initial fixture with the values for the site
    * The application `django-defaultsite <https://github.com/oppian/django-defaultsite>`_
    * Other options discussed `here <https://groups.google.com/forum/#!topic/django-developers/X-ef0C0V8Rk>`_
    """
    from django.contrib.sites import models as site_app
    site_fields = {'domain': settings.SITE_DOMAIN, 'name': settings.SITE_NAME}
    site = site_app.Site.objects.update_or_create(pk=settings.SITE_ID, defaults=site_fields)[0]
    logger.info(
        f'Updated settings of Site "{site.name}" with ID {site.pk} and domain {site.domain}')


def setup_postgresql_hstore_extension(sender, connection, **kwargs):
    from psycopg2.extras import register_hstore
    cursor = connection.connection.cursor()
    cursor.execute('CREATE EXTENSION IF NOT EXISTS hstore')
    register_hstore(connection.connection, globally=True)


def strip_strings_and_validate_model(sender, instance, raw, **kwargs):
    """Strip the string fields of the instance and run the instance's full_clean()."""
    if not raw:
        logger.debug(f'Validate model {instance} on save() with kwargs={kwargs}')
        for field in instance._meta.fields:
            if isinstance(field, (fields.CharField, fields.TextField)):
                field_value = getattr(instance, field.name)
                if isinstance(field_value, str):
                    setattr(instance, field.name, field_value.strip())
        instance.full_clean()



"""
Mix-ins for building your own forms.
"""

from django.forms import fields

from pytoolbox import module
from . import utils, widgets

_all = module.All(globals())


class ConvertEmailToTextMixin(object):
    """
    Set email inputs as text to avoid the i18n issue
    http://html5doctor.com/html5-forms-input-types#input-email.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            if getattr(field.widget, 'input_type', None) == 'email':
                field.widget.input_type = 'text'


class EnctypeMixin(object):

    @property
    def enctype(self):
        return 'multipart/form-data' if self.is_multipart() else 'application/x-www-form-urlencoded'


class HelpTextToPlaceholderMixin(object):
    """
    Update the widgets of the form to copy (and remove) the field's help text to the widget's
    placeholder.
    """

    #: Add a placeholder to the type of fields listed here.
    placeholder_fields = (
        fields.CharField,
        fields.DateField,
        fields.DateTimeField,
        fields.DecimalField,
        fields.EmailField,
        fields.FloatField,
        fields.IntegerField,
        fields.RegexField,
        fields.SlugField,
        fields.TimeField
    )
    #: Remove the help text after having copied it to the placeholder.
    placeholder_remove_help_text = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for name, field in self.fields.items():
            if field and isinstance(field, self.placeholder_fields):
                self.set_placeholder(name, field)

    def set_placeholder(self, name, field):  # pylint:disable=unused-argument
        field.widget.attrs['placeholder'] = field.help_text
        if self.placeholder_remove_help_text:
            field.help_text = None


class MapErrorsMixin(object):
    """
    Map errors based on field name. Mandatory when the form contains a field from a model named
    differently.
    """

    errors_map = {}

    def add_error(self, field, error):
        field = self.errors_map.get(field, field)
        return super(MapErrorsMixin, self).add_error(field, error)


class ModelBasedFormCleanupMixin(object):
    """
    Make possible the cleanup of the form by the model through a class method called `clean_form`.
    Useful to cleanup the form based on complex conditions, e.g. if two fields are inter-related
    (start/end dates, ...).
    """

    def clean(self):
        super().clean()
        try:
            return self._meta.model.clean_form(self)
        except AttributeError:
            return self.cleaned_data


class RequestMixin(object):
    """
    Accept request as a optional (default: None) argument of the constructor and set it as an
    attribute of the object.
    """

    def __init__(self, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super().__init__(*args, **kwargs)


class CreatedByMixin(RequestMixin):
    """Set instance's created_by field to current user if the instance is just created."""

    def save(self, commit=True):
        if hasattr(self.instance, 'created_by_id') and not self.instance.created_by_id:
            self.instance.created_by = self.request.user
        return super().save(commit=commit)


class StaffOnlyFieldsMixin(RequestMixin):
    """Hide some fields if authenticated user is not a member of the staff."""

    staff_only_fields = ()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.request or not self.request.user.is_staff:
            for field in self.staff_only_fields:
                self.fields.pop(field, None)


class UpdateWidgetAttributeMixin(object):
    """
    Update the widgets of the form based on a set of rules applied depending of the form field's
    class. The rules can change the class of the widget and/or update the attributes of the widget
    with :func:`pytoolbox.django.forms.utils.update_widget_attributes`.
    """

    #: Set of rules linking the form field's class to the replacement class and the attributes
    #  update list.
    widgets_rules = {
        fields.DateField: [widgets.CalendarDateInput, {'class': '+dateinput +input-small'}],
        fields.TimeField: [widgets.ClockTimeInput, {'class': '+timeinput +input-small'}],
    }
    #: Attributes that are applied to all widgets of the form
    widgets_common_attrs = {}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for field in self.fields.values():
            updates = self.widgets_rules.get(field.__class__)
            # May Update widget class with rules-based replacement class
            if updates and updates[0]:
                field.widget = updates[0]()
            # May update widget attributes with common attributes
            if self.widgets_common_attrs:
                utils.update_widget_attributes(field.widget, self.widgets_common_attrs)
            # May update widget attributes with rules-based attributes
            if updates and updates[1]:
                utils.update_widget_attributes(field.widget, updates[1])
        try:
            self._meta.model.init_form(self)
        except AttributeError:
            pass


__all__ = _all.diff(globals())
